# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import re
import stat
import tempfile

from mock import call, patch, mock_open
from oslo_concurrency.processutils import UnknownArgumentError
import six
from testtools import ExpectedException

from trove.common import exception
from trove.common.stream_codecs import (
    Base64Codec, IdentityCodec, IniCodec, JsonCodec,
    KeyValueCodec, PropertiesCodec, XmlCodec, YamlCodec)
from trove.common import utils
from trove.guestagent.common import guestagent_utils
from trove.guestagent.common import operating_system
from trove.guestagent.common.operating_system import FileMode
from trove.tests.unittests import trove_testtools


class TestOperatingSystem(trove_testtools.TestCase):

    def test_base64_codec(self):
        data = "Line 1\nLine 2\n"
        # Base64Codec.deserialize returns bytes instead of string.
        self._test_file_codec(data, Base64Codec(),
                              expected_data=data.encode('utf-8'))

        # when encoding is reversed for Base64Codec, reading from files
        # will call Base64Codec.serialize which returns string.
        data = "TGluZSAxCkxpbmUgMgo="
        self._test_file_codec(data, Base64Codec(), reverse_encoding=True)

        data = "5Am9+y0wTwqUx39sMMBg3611FWg="
        self._test_file_codec(data, Base64Codec(), reverse_encoding=True)

    def test_identity_file_codec(self):
        data = ("Lorem Ipsum, Lorem Ipsum\n"
                "Lorem Ipsum, Lorem Ipsum\n"
                "Lorem Ipsum, Lorem Ipsum\n")

        self._test_file_codec(data, IdentityCodec())

    def test_ini_file_codec(self):
        data_no_none = {"Section1": {"s1k1": 's1v1',
                                     "s1k2": 3.1415926535},
                        "Section2": {"s2k1": 1,
                                     "s2k2": True}}

        self._test_file_codec(data_no_none, IniCodec())

        data_with_none = {"Section1": {"s1k1": 's1v1',
                                       "s1k2": 3.1415926535},
                          "Section2": {"s2k1": 1,
                                       "s2k2": True,
                                       "s2k3": None}}

        # Keys with None values will be written without value.
        self._test_file_codec(data_with_none, IniCodec())

        # None will be replaced with 'default_value'.
        default_value = 1
        expected_data = guestagent_utils.update_dict(
            {"Section2": {"s2k3": default_value}}, dict(data_with_none))
        self._test_file_codec(data_with_none,
                              IniCodec(default_value=default_value),
                              expected_data=expected_data)

    def test_yaml_file_codec(self):
        data = {"Section1": 's1v1',
                "Section2": {"s2k1": '1',
                             "s2k2": 'True'},
                "Section3": {"Section4": {"s4k1": '3.1415926535',
                                          "s4k2": None}},
                "Section5": {"s5k1": 1,
                             "s5k2": True},
                "Section6": {"Section7": {"s7k1": 3.1415926535,
                                          "s7k2": None}}
                }

        self._test_file_codec(data, YamlCodec())
        self._test_file_codec(data, YamlCodec(default_flow_style=True))

    def test_properties_file_codec(self):
        data = {'key1': [1, "str1", '127.0.0.1', 3.1415926535, True, None],
                'key2': [2.0, 3, 0, "str1 str2"],
                'key3': ['str1', 'str2'],
                'key4': [],
                'key5': 5000,
                'key6': 'str1',
                'key7': 0,
                'key8': None,
                'key9': [['str1', 'str2'], ['str3', 'str4']],
                'key10': [['str1', 'str2', 'str3'], ['str3', 'str4'], 'str5'],
                'key11': True
                }

        self._test_file_codec(data, PropertiesCodec())
        self._test_file_codec(data, PropertiesCodec(
            string_mappings={'yes': True, 'no': False, "''": None}))

    def test_key_value_file_codec(self):
        data = {'key1': 'value1',
                'key2': 'value2',
                'key3': 'value3'}

        self._test_file_codec(data, KeyValueCodec())

    def test_json_file_codec(self):
        data = {"Section1": 's1v1',
                "Section2": {"s2k1": '1',
                             "s2k2": 'True'},
                "Section3": {"Section4": {"s4k1": '3.1415926535',
                                          "s4k2": None}},
                "Section5": {"s5k1": 1,
                             "s5k2": True},
                "Section6": {"Section7": {"s7k1": 3.1415926535,
                                          "s7k2": None}}
                }

        self._test_file_codec(data, JsonCodec())

    def test_xml_file_codec(self):
        data = {'document': {'@name': 'mydocument', '@ttl': '10',
                             'author': {'@name': 'Jycll ;-)'},
                             'page': [{'@number': '1', 'paragraph':
                                       ['lorem ipsum', 'more lorem ipsum']},
                                      {'@number': '1', 'paragraph':
                                       ['lorem ipsum', 'more lorem ipsum']}]
                             }
                }

        self._test_file_codec(data, XmlCodec())

    def _test_file_codec(self, data, read_codec, write_codec=None,
                         expected_data=None,
                         expected_exception=None,
                         reverse_encoding=False):
        write_codec = write_codec or read_codec

        with tempfile.NamedTemporaryFile() as test_file:
            encode = True
            decode = True
            if reverse_encoding:
                encode = False
                decode = False
            if expected_exception:
                with expected_exception:
                    operating_system.write_file(test_file.name, data,
                                                codec=write_codec,
                                                encode=encode)
                    operating_system.read_file(test_file.name,
                                               codec=read_codec,
                                               decode=decode)
            else:
                operating_system.write_file(test_file.name, data,
                                            codec=write_codec,
                                            encode=encode)
                read = operating_system.read_file(test_file.name,
                                                  codec=read_codec,
                                                  decode=decode)
                if expected_data is not None:
                    self.assertEqual(expected_data, read)
                else:
                    self.assertEqual(data, read)

    def test_read_write_file_input_validation(self):
        with ExpectedException(exception.UnprocessableEntity,
                               "File does not exist: None"):
            operating_system.read_file(None)

        with ExpectedException(exception.UnprocessableEntity,
                               "File does not exist: /__DOES_NOT_EXIST__"):
            operating_system.read_file('/__DOES_NOT_EXIST__')

        with ExpectedException(exception.UnprocessableEntity,
                               "Invalid path: None"):
            operating_system.write_file(None, {})

    @patch.object(operating_system, 'copy')
    def test_write_file_as_root(self, copy_mock):
        target_file = tempfile.NamedTemporaryFile()
        temp_file = tempfile.NamedTemporaryFile('w')

        with patch('tempfile.NamedTemporaryFile', return_value=temp_file):
            operating_system.write_file(
                target_file.name, "Lorem Ipsum", as_root=True)
            copy_mock.assert_called_once_with(
                temp_file.name, target_file.name, force=True, as_root=True)
        self.assertFalse(os.path.exists(temp_file.name))

    @patch.object(operating_system, 'copy',
                  side_effect=Exception("Error while executing 'copy'."))
    def test_write_file_as_root_with_error(self, copy_mock):
        target_file = tempfile.NamedTemporaryFile()
        temp_file = tempfile.NamedTemporaryFile('w')
        with patch('tempfile.NamedTemporaryFile', return_value=temp_file):
            with ExpectedException(Exception, "Error while executing 'copy'."):
                operating_system.write_file(target_file.name,
                                            "Lorem Ipsum", as_root=True)
        self.assertFalse(os.path.exists(temp_file.name))

    @patch.object(operating_system, 'exists', return_value=True)
    @patch.object(operating_system, 'copy')
    @patch.object(operating_system, 'chmod')
    @patch.object(IdentityCodec, 'deserialize')
    @patch.object(IdentityCodec, 'serialize')
    @patch.object(operating_system, 'open',
                  mock_open(read_data='MockingRead'))
    def test_read_file_with_flags_and_conv_func(self, mock_serialize,
                                                mock_deserialize,
                                                mock_chmod, mock_copy,
                                                *args):
        test_path = '/path/of/file'
        test_data = 'MockingRead'
        # use getattr to avoid pylint 'no-member' warning
        mock_file = getattr(operating_system, 'open')

        # simple read
        operating_system.read_file(test_path)
        mock_file.assert_called_once_with(test_path, 'r')
        mock_file().read.assert_called_once()
        mock_deserialize.called_once_with(test_data)
        mock_file.reset_mock()
        mock_deserialize.reset_mock()

        # read with decode=False
        operating_system.read_file(test_path, decode=False)
        mock_file.assert_called_once_with(test_path, 'rb')
        mock_file().read.assert_called_once()
        mock_serialize.called_once_with(test_data)
        mock_file.reset_mock()
        mock_serialize.reset_mock()

        # checking _read_file_as_root arguments
        with patch.object(operating_system,
                          '_read_file_as_root') as mock_read_file_as_root:
            # simple read as root,
            operating_system.read_file(test_path, as_root=True)
            mock_read_file_as_root.assert_called_once_with(
                test_path, 'r', mock_deserialize)
            mock_deserialize.assert_not_called()
            mock_read_file_as_root.reset_mock()

            # read as root with decode=False,
            operating_system.read_file(test_path, as_root=True, decode=False)
            mock_read_file_as_root.assert_called_once_with(
                test_path, 'rb', mock_serialize)
            mock_serialize.assert_not_called()

        # simple read as root
        temp_file = tempfile.NamedTemporaryFile('r')
        with patch.object(tempfile, 'NamedTemporaryFile',
                          return_value=temp_file) as mock_temp_file:
            operating_system.read_file(test_path, as_root=True)
            mock_temp_file.assert_called_once_with('r')
            mock_copy.called_once_with(test_path, temp_file.name,
                                       force=True, dereference=True,
                                       as_root=True)
            mock_chmod.called_once_with(temp_file.name,
                                        FileMode.ADD_READ_ALL(),
                                        as_root=True)
            mock_deserialize.assert_called_once_with('')
        self.assertFalse(os.path.exists(temp_file.name))
        mock_copy.reset_mock()
        mock_chmod.reset_mock()
        mock_deserialize.reset_mock()

        # read as root with decode=False
        temp_file = tempfile.NamedTemporaryFile('rb')
        with patch.object(tempfile, 'NamedTemporaryFile',
                          return_value=temp_file) as mock_temp_file:
            operating_system.read_file(test_path, as_root=True,
                                       decode=False)
            mock_temp_file.assert_called_once_with('rb')
            mock_copy.called_once_with(test_path, temp_file.name,
                                       force=True, dereference=True,
                                       as_root=True)
            mock_chmod.called_once_with(temp_file.name,
                                        FileMode.ADD_READ_ALL(),
                                        as_root=True)
            mock_serialize.assert_called_once_with(b'')
        self.assertFalse(os.path.exists(temp_file.name))

    @patch.object(operating_system, 'copy')
    @patch.object(operating_system, 'chmod')
    @patch.object(IdentityCodec, 'deserialize',
                  return_value=b'DeseiralizedData')
    @patch.object(IdentityCodec, 'serialize',
                  return_value='SerializedData')
    @patch.object(operating_system, 'open', mock_open())
    def test_write_file_with_flags_and_conv_func(self, mock_serialize,
                                                 mock_deserialize,
                                                 mock_chmod, mock_copy):
        test_path = '/path/of/file'
        test_data = 'MockingWrite'
        test_serialize = 'SerializedData'
        test_deserialize = b'DeseiralizedData'
        mock_file = getattr(operating_system, 'open')

        # simple write
        operating_system.write_file(test_path, test_data)
        mock_file.assert_called_once_with(test_path, 'w')
        mock_serialize.called_once_with(test_data)
        mock_file().write.assert_called_once_with(test_serialize)
        mock_file().flush.assert_called_once()
        mock_file.reset_mock()
        mock_serialize.reset_mock()

        # write with encode=False
        operating_system.write_file(test_path, test_data, encode=False)
        mock_file.assert_called_once_with(test_path, 'wb')
        mock_deserialize.called_once_with(test_data)
        mock_file().write.assert_called_once_with(test_deserialize)
        mock_file().flush.assert_called_once()
        mock_file.reset_mock()
        mock_deserialize.reset_mock()

        # checking _write_file_as_root arguments
        with patch.object(operating_system,
                          '_write_file_as_root') as mock_write_file_as_root:
            # simple write as root,
            operating_system.write_file(test_path, test_data, as_root=True)
            mock_write_file_as_root.assert_called_once_with(
                test_path, test_data, 'w', mock_serialize)
            mock_serialize.assert_not_called()
            mock_write_file_as_root.reset_mock()

            # read as root with encode=False,
            operating_system.write_file(test_path, test_data,
                                        as_root=True, encode=False)
            mock_write_file_as_root.assert_called_once_with(
                test_path, test_data, 'wb', mock_deserialize)
            mock_deserialize.assert_not_called()

        # simple write as root
        temp_file = tempfile.NamedTemporaryFile('w')
        with patch.object(tempfile, 'NamedTemporaryFile',
                          return_value=temp_file) as mock_temp_file:
            operating_system.write_file(test_path, test_data, as_root=True)
            mock_temp_file.assert_called_once_with('w', delete=False)
            mock_serialize.assert_called_once_with(test_data)
            mock_copy.called_once_with(temp_file.name, test_path,
                                       force=True, as_root=True)
        self.assertFalse(os.path.exists(temp_file.name))
        mock_copy.reset_mock()
        mock_chmod.reset_mock()
        mock_serialize.reset_mock()

        # write as root with decode=False
        temp_file = tempfile.NamedTemporaryFile('wb')
        with patch.object(tempfile, 'NamedTemporaryFile',
                          return_value=temp_file) as mock_temp_file:
            operating_system.write_file(test_path, test_data,
                                        as_root=True, encode=False)
            mock_temp_file.assert_called_once_with('wb', delete=False)
            mock_deserialize.assert_called_once_with(test_data)
            mock_copy.called_once_with(temp_file.name, test_path,
                                       force=True, as_root=True)
        self.assertFalse(os.path.exists(temp_file.name))

    def test_start_service(self):
        self._assert_service_call(operating_system.start_service,
                                  'cmd_start')

    def test_stop_service(self):
        self._assert_service_call(operating_system.stop_service,
                                  'cmd_stop')

    def test_enable_service_on_boot(self):
        self._assert_service_call(operating_system.enable_service_on_boot,
                                  'cmd_enable')

    def test_disable_service_on_boot(self):
        self._assert_service_call(operating_system.disable_service_on_boot,
                                  'cmd_disable')

    @patch.object(operating_system, '_execute_service_command')
    def _assert_service_call(self, fun, expected_cmd_key,
                             exec_service_cmd_mock):
        test_candidate_names = ['test_service_1', 'test_service_2']
        fun(test_candidate_names)
        exec_service_cmd_mock.assert_called_once_with(test_candidate_names,
                                                      expected_cmd_key)

    @patch.object(operating_system, 'service_discovery',
                  return_value={'cmd_start': 'start',
                                'cmd_stop': 'stop',
                                'cmd_enable': 'enable',
                                'cmd_disable': 'disable'})
    def test_execute_service_command(self, discovery_mock):
        test_service_candidates = ['service_name']
        self._assert_execute_call([['start']], [{'shell': True}],
                                  operating_system._execute_service_command,
                                  None, test_service_candidates, 'cmd_start')
        discovery_mock.assert_called_once_with(test_service_candidates)

        with ExpectedException(exception.UnprocessableEntity,
                               "Candidate service names not specified."):
            operating_system._execute_service_command([], 'cmd_disable')

        with ExpectedException(exception.UnprocessableEntity,
                               "Candidate service names not specified."):
            operating_system._execute_service_command(None, 'cmd_start')

        with ExpectedException(RuntimeError, "Service control command not "
                               "available: unknown"):
            operating_system._execute_service_command(test_service_candidates,
                                                      'unknown')

    def test_modes(self):
        self._assert_modes(None, None, None, operating_system.FileMode())
        self._assert_modes(None, None, None,
                           operating_system.FileMode([], [], []))
        self._assert_modes(0o770, 0o4, 0o3, operating_system.FileMode(
            [stat.S_IRWXU, stat.S_IRWXG],
            [stat.S_IROTH],
            [stat.S_IWOTH | stat.S_IXOTH])
        )
        self._assert_modes(0o777, None, None, operating_system.FileMode(
            [stat.S_IRWXU, stat.S_IRWXG, stat.S_IRWXO])
        )
        self._assert_modes(0o777, None, None, operating_system.FileMode(
            reset=[stat.S_IRWXU, stat.S_IRWXG, stat.S_IRWXO])
        )
        self._assert_modes(None, 0o777, None, operating_system.FileMode(
            add=[stat.S_IRWXU, stat.S_IRWXG, stat.S_IRWXO])
        )
        self._assert_modes(None, None, 0o777, operating_system.FileMode(
            remove=[stat.S_IRWXU, stat.S_IRWXG, stat.S_IRWXO])
        )

        self.assertEqual(
            operating_system.FileMode(add=[stat.S_IRUSR, stat.S_IWUSR]),
            operating_system.FileMode(add=[stat.S_IWUSR, stat.S_IRUSR]))

        self.assertEqual(
            hash(operating_system.FileMode(add=[stat.S_IRUSR, stat.S_IWUSR])),
            hash(operating_system.FileMode(add=[stat.S_IWUSR, stat.S_IRUSR])))

        self.assertNotEqual(
            operating_system.FileMode(add=[stat.S_IRUSR, stat.S_IWUSR]),
            operating_system.FileMode(reset=[stat.S_IRUSR, stat.S_IWUSR]))

        self.assertNotEqual(
            hash(operating_system.FileMode(add=[stat.S_IRUSR, stat.S_IWUSR])),
            hash(operating_system.FileMode(reset=[stat.S_IRUSR, stat.S_IWUSR]))
        )

    def _assert_modes(self, ex_reset, ex_add, ex_remove, actual):
        self.assertEqual(bool(ex_reset or ex_add or ex_remove),
                         actual.has_any())
        self.assertEqual(ex_reset, actual.get_reset_mode())
        self.assertEqual(ex_add, actual.get_add_mode())
        self.assertEqual(ex_remove, actual.get_remove_mode())

    def test_chmod(self):
        self._assert_execute_call(
            [['chmod', '-R', '=064', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chmod, None,
            'path', FileMode.SET_GRP_RW_OTH_R,
            as_root=True)
        self._assert_execute_call(
            [['chmod', '-R', '+444', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chmod, None,
            'path', FileMode.ADD_READ_ALL,
            as_root=True)

        self._assert_execute_call(
            [['chmod', '-R', '+060', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chmod, None,
            'path', FileMode.ADD_GRP_RW,
            as_root=True)

        self._assert_execute_call(
            [['chmod', '-R', '=777', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chmod, None,
            'path', FileMode.SET_FULL,
            as_root=True)

        self._assert_execute_call(
            [['chmod', '-f', '=777', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chmod, None,
            'path', FileMode.SET_FULL,
            as_root=True, recursive=False, force=True)

        self._assert_execute_call(
            [['chmod', '-R', '=777', 'path']],
            [{'timeout': 100}],
            operating_system.chmod, None,
            'path', FileMode.SET_FULL,
            timeout=100)

        self._assert_execute_call(
            [['chmod', '-R', '=777', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo', 'timeout': None}],
            operating_system.chmod, None,
            'path', FileMode.SET_FULL,
            as_root=True, timeout=None)

        self._assert_execute_call(
            None, None,
            operating_system.chmod,
            ExpectedException(exception.UnprocessableEntity,
                              "No file mode specified."),
            'path', FileMode())

        self._assert_execute_call(
            None, None,
            operating_system.chmod,
            ExpectedException(exception.UnprocessableEntity,
                              "No file mode specified."),
            'path', None)

        self._assert_execute_call(
            None, None,
            operating_system.chmod,
            ExpectedException(exception.UnprocessableEntity,
                              "Cannot change mode of a blank file."),
            '', FileMode.SET_FULL)

        self._assert_execute_call(
            None, None,
            operating_system.chmod,
            ExpectedException(exception.UnprocessableEntity,
                              "Cannot change mode of a blank file."),
            None, FileMode.SET_FULL)

        self._assert_execute_call(
            None, None,
            operating_system.chmod,
            ExpectedException(UnknownArgumentError,
                              "Got unknown keyword args: {'_unknown_kw': 0}"),
            'path', FileMode.SET_FULL, _unknown_kw=0)

    def test_remove(self):
        self._assert_execute_call(
            [['rm', '-R', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.remove, None, 'path', as_root=True)

        self._assert_execute_call(
            [['rm', '-f', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.remove, None, 'path', recursive=False, force=True,
            as_root=True)

        self._assert_execute_call(
            [['rm', '-R', 'path']],
            [{'timeout': 100}],
            operating_system.remove, None,
            'path', timeout=100)

        self._assert_execute_call(
            [['rm', '-R', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo', 'timeout': None}],
            operating_system.remove, None, 'path', timeout=None, as_root=True)

        self._assert_execute_call(
            None, None,
            operating_system.remove,
            ExpectedException(exception.UnprocessableEntity,
                              "Cannot remove a blank file."), '')

        self._assert_execute_call(
            None, None,
            operating_system.remove,
            ExpectedException(exception.UnprocessableEntity,
                              "Cannot remove a blank file."), None)

        self._assert_execute_call(
            None, None,
            operating_system.remove,
            ExpectedException(UnknownArgumentError,
                              "Got unknown keyword args: {'_unknown_kw': 0}"),
            'path', _unknown_kw=0)

    def test_move(self):
        self._assert_execute_call(
            [['mv', 'source', 'destination']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.move, None, 'source', 'destination', as_root=True)

        self._assert_execute_call(
            [['mv', '-f', 'source', 'destination']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.move, None, 'source', 'destination', force=True,
            as_root=True)

        self._assert_execute_call(
            [['mv', 'source', 'destination']],
            [{'timeout': 100}],
            operating_system.move, None, 'source', 'destination',
            timeout=100)

        self._assert_execute_call(
            [['mv', 'source', 'destination']],
            [{'run_as_root': True, 'root_helper': 'sudo', 'timeout': None}],
            operating_system.move, None, 'source', 'destination', timeout=None,
            as_root=True)

        self._assert_execute_call(
            None, None,
            operating_system.move,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing source path."), '', 'destination')

        self._assert_execute_call(
            None, None,
            operating_system.move,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing source path."), None, 'destination')

        self._assert_execute_call(
            None, None,
            operating_system.move,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing destination path."), 'source', '')

        self._assert_execute_call(
            None, None,
            operating_system.move,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing destination path."), 'source', None)

        self._assert_execute_call(
            None, None,
            operating_system.move,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing source path."), '', '')

        self._assert_execute_call(
            None, None,
            operating_system.move,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing source path."), None, None)

        self._assert_execute_call(
            None, None,
            operating_system.move,
            ExpectedException(UnknownArgumentError,
                              "Got unknown keyword args: {'_unknown_kw': 0}"),
            'source', 'destination', _unknown_kw=0)

    def test_copy(self):
        self._assert_execute_call(
            [['cp', '-R', 'source', 'destination']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.copy, None, 'source', 'destination', as_root=True)

        self._assert_execute_call(
            [['cp', '-f', '-p', 'source', 'destination']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.copy, None, 'source', 'destination', force=True,
            preserve=True, recursive=False, as_root=True)

        self._assert_execute_call(
            [['cp', '-R', 'source', 'destination']],
            [{'timeout': 100}],
            operating_system.copy, None, 'source', 'destination',
            timeout=100)

        self._assert_execute_call(
            [['cp', '-R', 'source', 'destination']],
            [{'run_as_root': True, 'root_helper': "sudo", 'timeout': None}],
            operating_system.copy, None, 'source', 'destination', timeout=None,
            as_root=True)

        self._assert_execute_call(
            None, None,
            operating_system.copy,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing source path."), '', 'destination')

        self._assert_execute_call(
            None, None,
            operating_system.copy,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing source path."), None, 'destination')

        self._assert_execute_call(
            None, None,
            operating_system.copy,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing destination path."), 'source', '')

        self._assert_execute_call(
            None, None,
            operating_system.copy,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing destination path."), 'source', None)

        self._assert_execute_call(
            None, None,
            operating_system.copy,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing source path."), '', '')

        self._assert_execute_call(
            None, None,
            operating_system.copy,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing source path."), None, None)

        self._assert_execute_call(
            None, None,
            operating_system.copy,
            ExpectedException(UnknownArgumentError,
                              "Got unknown keyword args: {'_unknown_kw': 0}"),
            'source', 'destination', _unknown_kw=0)

    def test_chown(self):
        self._assert_execute_call(
            [['chown', '-R', 'usr:grp', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chown, None, 'path', 'usr', 'grp', as_root=True)

        self._assert_execute_call(
            [['chown', 'usr:grp', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chown, None,
            'path', 'usr', 'grp', recursive=False, as_root=True)

        self._assert_execute_call(
            [['chown', '-f', '-R', 'usr:grp', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chown, None,
            'path', 'usr', 'grp', force=True, as_root=True)

        self._assert_execute_call(
            [['chown', '-R', ':grp', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chown, None, 'path', '', 'grp', as_root=True)

        self._assert_execute_call(
            [['chown', '-R', 'usr:', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chown, None, 'path', 'usr', '', as_root=True)

        self._assert_execute_call(
            [['chown', '-R', ':grp', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chown, None, 'path', None, 'grp', as_root=True)

        self._assert_execute_call(
            [['chown', '-R', 'usr:', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.chown, None, 'path', 'usr', None, as_root=True)

        self._assert_execute_call(
            [['chown', '-R', 'usr:', 'path']],
            [{'timeout': 100}],
            operating_system.chown, None,
            'path', 'usr', None, timeout=100)

        self._assert_execute_call(
            [['chown', '-R', 'usr:', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo',
              'timeout': None}],
            operating_system.chown, None,
            'path', 'usr', None, timeout=None, as_root=True)

        self._assert_execute_call(
            None, None,
            operating_system.chown,
            ExpectedException(exception.UnprocessableEntity,
                              "Cannot change ownership of a blank file."),
            '', 'usr', 'grp')

        self._assert_execute_call(
            None, None,
            operating_system.chown,
            ExpectedException(exception.UnprocessableEntity,
                              "Cannot change ownership of a blank file."),
            None, 'usr', 'grp')

        self._assert_execute_call(
            None, None,
            operating_system.chown,
            ExpectedException(exception.UnprocessableEntity,
                              "Please specify owner or group, or both."),
            'path', '', '')

        self._assert_execute_call(
            None, None,
            operating_system.chown,
            ExpectedException(exception.UnprocessableEntity,
                              "Please specify owner or group, or both."),
            'path', None, None)

        self._assert_execute_call(
            None, None,
            operating_system.chown,
            ExpectedException(exception.UnprocessableEntity,
                              "Cannot change ownership of a blank file."),
            None, None, None)

        self._assert_execute_call(
            None, None,
            operating_system.chown,
            ExpectedException(exception.UnprocessableEntity,
                              "Cannot change ownership of a blank file."),
            '', '', '')

        self._assert_execute_call(
            None, None,
            operating_system.chown,
            ExpectedException(UnknownArgumentError,
                              "Got unknown keyword args: {'_unknown_kw': 0}"),
            'path', 'usr', None, _unknown_kw=0)

    def test_change_user_group(self):
        self._assert_execute_call(
            [['usermod', '-a', '-G', 'user', 'group']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.change_user_group, None, 'group', 'user',
            as_root=True)

        self._assert_execute_call(
            [['usermod', '-a', '-G', 'user', 'group']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.change_user_group, None, 'group', 'user',
            append=True, add_group=True, as_root=True)

        self._assert_execute_call(
            [['usermod', '-a', '-G', 'user', 'group']],
            [{'timeout': 100}],
            operating_system.change_user_group, None, 'group', 'user',
            timeout=100)

        self._assert_execute_call(
            [['usermod', '-a', '-G', 'user', 'group']],
            [{'run_as_root': True, 'root_helper': "sudo", 'timeout': None}],
            operating_system.change_user_group, None, 'group', 'user',
            timeout=None, as_root=True)

        self._assert_execute_call(
            None, None,
            operating_system.change_user_group,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing user."), '', 'group')

        self._assert_execute_call(
            None, None,
            operating_system.change_user_group,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing user."), None, 'group')

        self._assert_execute_call(
            None, None,
            operating_system.change_user_group,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing group."), 'user', '')

        self._assert_execute_call(
            None, None,
            operating_system.change_user_group,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing group."), 'user', None)

        self._assert_execute_call(
            None, None,
            operating_system.change_user_group,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing user."), '', '')

        self._assert_execute_call(
            None, None,
            operating_system.change_user_group,
            ExpectedException(exception.UnprocessableEntity,
                              "Missing user."), None, None)

        self._assert_execute_call(
            None, None,
            operating_system.change_user_group,
            ExpectedException(UnknownArgumentError,
                              "Got unknown keyword args: {'_unknown_kw': 0}"),
            'user', 'add_group', _unknown_kw=0)

    def test_create_directory(self):
        self._assert_execute_call(
            [['mkdir', '-p', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.create_directory, None, 'path', as_root=True)

        self._assert_execute_call(
            [['mkdir', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.create_directory, None, 'path', force=False,
            as_root=True)

        self._assert_execute_call(
            [['mkdir', '-p', 'path'], ['chown', '-R', 'usr:grp', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'},
             {'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.create_directory, None,
            'path', user='usr', group='grp', as_root=True)

        self._assert_execute_call(
            [['mkdir', '-p', 'path'], ['chown', '-R', ':grp', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'},
             {'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.create_directory, None, 'path', group='grp',
            as_root=True)

        self._assert_execute_call(
            [['mkdir', '-p', 'path'], ['chown', '-R', 'usr:', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'},
             {'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.create_directory, None, 'path', user='usr',
            as_root=True)

        self._assert_execute_call(
            [['mkdir', '-p', 'path'], ['chown', '-R', 'usr:', 'path']],
            [{'timeout': 100}, {'timeout': 100}],
            operating_system.create_directory, None,
            'path', user='usr', timeout=100)

        self._assert_execute_call(
            [['mkdir', '-p', 'path'], ['chown', '-R', 'usr:', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo', 'timeout': None},
             {'run_as_root': True, 'root_helper': 'sudo', 'timeout': None}],
            operating_system.create_directory, None,
            'path', user='usr', timeout=None, as_root=True)

        self._assert_execute_call(
            [['mkdir', '-p', 'path'], ['chown', '-R', 'usr:', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'},
             {'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.create_directory, None,
            'path', user='usr', group='', as_root=True)

        self._assert_execute_call(
            [['mkdir', '-p', 'path'], ['chown', '-R', ':grp', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'},
             {'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.create_directory, None,
            'path', user='', group='grp', as_root=True)

        self._assert_execute_call(
            [['mkdir', '-p', 'path']],
            [{'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.create_directory, None, 'path', user='', group='',
            as_root=True)

        self._assert_execute_call(
            None, None,
            operating_system.create_directory,
            ExpectedException(exception.UnprocessableEntity,
                              "Cannot create a blank directory."),
            '', user='usr', group='grp')

        self._assert_execute_call(
            None, None,
            operating_system.create_directory,
            ExpectedException(exception.UnprocessableEntity,
                              "Cannot create a blank directory."), None)

        self._assert_execute_call(
            None, None,
            operating_system.create_directory,
            ExpectedException(UnknownArgumentError,
                              "Got unknown keyword args: {'_unknown_kw': 0}"),
            'path', _unknown_kw=0)

    def test_find_executable_without_path(self):
        command = "command"
        self._delegate_assert_find_executable(command=command,
                                              path=None,
                                              isfile=True,
                                              access=True,
                                              expected_return_value=(
                                                  "/usr/bin/command"))
        self._delegate_assert_find_executable(command=command,
                                              path=None,
                                              isfile=True,
                                              access=False,
                                              expected_return_value=None)
        self._delegate_assert_find_executable(command=command,
                                              path=None,
                                              isfile=False,
                                              access=True,
                                              expected_return_value=None)
        self._delegate_assert_find_executable(command=command,
                                              path=None,
                                              isfile=False,
                                              access=False,
                                              expected_return_value=None)

    def test_find_executable_with_path(self):
        command = "command"
        path = "/home"
        self._delegate_assert_find_executable(command=command,
                                              path=path,
                                              isfile=True,
                                              access=True,
                                              expected_return_value=(
                                                  "/home/command"))
        self._delegate_assert_find_executable(command=command,
                                              path=path,
                                              isfile=True,
                                              access=False,
                                              expected_return_value=None)
        self._delegate_assert_find_executable(command=command,
                                              path=path,
                                              isfile=False,
                                              access=True,
                                              expected_return_value=None)
        self._delegate_assert_find_executable(command=command,
                                              path=path,
                                              isfile=False,
                                              access=False,
                                              expected_return_value=None)

    def _delegate_assert_find_executable(self, command, path, isfile,
                                         access, expected_return_value):
        self._assert_find_executable(command, path, isfile, access,
                                     expected_return_value)

    @patch.object(os, 'access')
    @patch.object(os.path, 'isfile')
    @patch.object(os.environ, 'get', return_value="/usr/bin")
    def _assert_find_executable(self, command, path, isfile, access,
                                expected_return_value, mock_environ,
                                mock_isfile, mock_access):
        mock_access.return_value = access
        mock_isfile.return_value = isfile
        actual_result = operating_system.find_executable(command, path)
        self.assertEqual(expected_return_value, actual_result)
        if path is None:
            mock_environ.assert_called_once()
        else:
            mock_environ.assert_not_called()

    def test_exists(self):
        self.assertFalse(
            operating_system.exists(tempfile.gettempdir(), is_directory=False))
        self.assertTrue(
            operating_system.exists(tempfile.gettempdir(), is_directory=True))

        with tempfile.NamedTemporaryFile() as test_file:
            self.assertTrue(
                operating_system.exists(test_file.name, is_directory=False))
            self.assertFalse(
                operating_system.exists(test_file.name, is_directory=True))

        self._assert_execute_call(
            [['test -f path && echo 1 || echo 0']],
            [{'shell': True, 'check_exit_code': False,
              'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.exists, None, 'path', is_directory=False,
            as_root=True)
        self._assert_execute_call(
            [['test -d path && echo 1 || echo 0']],
            [{'shell': True, 'check_exit_code': False,
              'run_as_root': True, 'root_helper': 'sudo'}],
            operating_system.exists, None, 'path', is_directory=True,
            as_root=True)

    def _assert_execute_call(self, exec_args, exec_kwargs,
                             func, return_value, *args, **kwargs):
        """
        Execute a function with given arguments.
        Assert a return value and appropriate sequence of calls to the
        'utils.execute_with_timeout' interface as the result.

        :param exec_args:         Expected arguments to the execute calls.
                                  This is a list-of-list where each sub-list
                                  represent a single call to
                                  'utils.execute_with_timeout'.
        :type exec_args:          list-of-lists

        :param exec_kwargs:       Expected keywords to the execute call.
                                  This is a list-of-dicts where each dict
                                  represent a single call to
                                  'utils.execute_with_timeout'.
        :type exec_kwargs:        list-of-dicts

        :param func:              Tested function call.
        :type func:               callable

        :param return_value:      Expected return value or exception
                                  from the tested call if any.
        :type return_value:       object

        :param args:              Arguments passed to the tested call.
        :type args:               list

        :param kwargs:            Keywords passed to the tested call.
        :type kwargs:             dict
        """

        with patch.object(utils, 'execute_with_timeout',
                          return_value=('0', '')) as exec_call:
            if isinstance(return_value, ExpectedException):
                with return_value:
                    func(*args, **kwargs)
            else:
                actual_value = func(*args, **kwargs)
                if return_value is not None:
                    self.assertEqual(return_value, actual_value,
                                     "Return value mismatch.")
                expected_calls = []
                for arg, kw in six.moves.zip(exec_args, exec_kwargs):
                    expected_calls.append(call(*arg, **kw))

                self.assertEqual(expected_calls, exec_call.mock_calls,
                                 "Mismatch in calls to "
                                 "'execute_with_timeout'.")

    def test_get_os_redhat(self):
        with patch.object(os.path, 'isfile', side_effect=[True]):
            find_os = operating_system.get_os()
        self.assertEqual('redhat', find_os)

    def test_get_os_suse(self):
        with patch.object(os.path, 'isfile', side_effect=[False, True]):
            find_os = operating_system.get_os()
        self.assertEqual('suse', find_os)

    def test_get_os_debian(self):
        with patch.object(os.path, 'isfile', side_effect=[False, False]):
            find_os = operating_system.get_os()
        self.assertEqual('debian', find_os)

    def test_upstart_type_service_discovery(self):
        with patch.object(os.path, 'isfile', side_effect=[True]):
            mysql_service = operating_system.service_discovery(["mysql"])
        self.assertIsNotNone(mysql_service['cmd_start'])
        self.assertIsNotNone(mysql_service['cmd_enable'])

    def test_sysvinit_type_service_discovery(self):
        with patch.object(os.path, 'isfile', side_effect=[False, True, True]):
            mysql_service = operating_system.service_discovery(["mysql"])
        self.assertIsNotNone(mysql_service['cmd_start'])
        self.assertIsNotNone(mysql_service['cmd_enable'])

    def test_sysvinit_chkconfig_type_service_discovery(self):
        with patch.object(os.path, 'isfile',
                          side_effect=[False, True, False, True]):
            mysql_service = operating_system.service_discovery(["mysql"])
        self.assertIsNotNone(mysql_service['cmd_start'])
        self.assertIsNotNone(mysql_service['cmd_enable'])

    @patch.object(os.path, 'islink', return_value=True)
    @patch.object(os.path, 'realpath')
    @patch.object(os.path, 'basename')
    def test_systemd_symlinked_type_service_discovery(self, mock_base,
                                                      mock_path, mock_islink):
        with patch.object(os.path, 'isfile', side_effect=[False, False, True]):
            mysql_service = operating_system.service_discovery(["mysql"])
        self.assertIsNotNone(mysql_service['cmd_start'])
        self.assertIsNotNone(mysql_service['cmd_enable'])

    def test_systemd_not_symlinked_type_service_discovery(self):
        with patch.object(os.path, 'isfile', side_effect=[False, False, True]):
            with patch.object(os.path, 'islink', return_value=False):
                mysql_service = operating_system.service_discovery(["mysql"])
        self.assertIsNotNone(mysql_service['cmd_start'])
        self.assertIsNotNone(mysql_service['cmd_enable'])

    def test_file_discovery(self):
        with patch.object(os.path, 'isfile', side_effect=[False, True]):
            config_file = operating_system.file_discovery(
                ["/etc/mongodb.conf", "/etc/mongod.conf"])
        self.assertEqual('/etc/mongod.conf', config_file)
        with patch.object(os.path, 'isfile', side_effect=[False]):
            config_file = operating_system.file_discovery(
                ["/etc/mongodb.conf"])
        self.assertEqual('', config_file)

    def test_list_files_in_directory(self):
        root_path = tempfile.mkdtemp()
        try:
            all_paths = set()
            self._create_temp_fs_structure(
                root_path, 3, 3, ['txt', 'py', ''], 1, all_paths)

            # All files in the top directory.
            self._assert_list_files(
                root_path, False, None, False, all_paths, 9)

            # All files & directories in the top directory.
            self._assert_list_files(
                root_path, False, None, True, all_paths, 10)

            # All files recursive.
            self._assert_list_files(
                root_path, True, None, False, all_paths, 27)

            # All files & directories recursive.
            self._assert_list_files(
                root_path, True, None, True, all_paths, 29)

            # Only '*.txt' in the top directory.
            self._assert_list_files(
                root_path, False, r'.*\.txt$', False, all_paths, 3)

            # Only '*.txt' (including directories) in the top directory.
            self._assert_list_files(
                root_path, False, r'.*\.txt$', True, all_paths, 3)

            # Only '*.txt' recursive.
            self._assert_list_files(
                root_path, True, r'.*\.txt$', True, all_paths, 9)

            # Only '*.txt' (including directories) recursive.
            self._assert_list_files(
                root_path, True, r'.*\.txt$', False, all_paths, 9)

            # Only extension-less files in the top directory.
            self._assert_list_files(
                root_path, False, r'[^\.]*$', False, all_paths, 3)

            # Only extension-less files recursive.
            self._assert_list_files(
                root_path, True, r'[^\.]*$', False, all_paths, 9)

            # Non-existing extension in the top directory.
            self._assert_list_files(
                root_path, False, r'.*\.bak$', False, all_paths, 0)

            # Non-existing extension recursive.
            self._assert_list_files(
                root_path, True, r'.*\.bak$', False, all_paths, 0)
        finally:
            try:
                os.remove(root_path)
            except Exception:
                pass  # Do not fail in the cleanup.

    def _assert_list_files(self, root, recursive, pattern, include_dirs,
                           all_paths, count):
        found = operating_system.list_files_in_directory(
            root, recursive=recursive, pattern=pattern,
            include_dirs=include_dirs)
        expected = {
            path for path in filter(
                lambda item: include_dirs or not os.path.isdir(item),
                all_paths) if (
                (recursive or os.path.dirname(path) == root) and (
                    not pattern or re.match(
                        pattern, os.path.basename(path))))}
        self.assertEqual(expected, found)
        self.assertEqual(count, len(found),
                         "Incorrect number of listed files.")

    def _create_temp_fs_structure(self, root_path,
                                  num_levels, num_files_per_extension,
                                  file_extensions, level, created_paths):
        """Create a structure of temporary directories 'num_levels' deep with
        temporary files on each level.
        """
        file_paths = self._create_temp_files(
            root_path, num_files_per_extension, file_extensions)
        created_paths.update(file_paths)

        if level < num_levels:
            path = tempfile.mkdtemp(dir=root_path)
            created_paths.add(path)
            self._create_temp_fs_structure(
                path, num_levels, num_files_per_extension,
                file_extensions, level + 1, created_paths)

    def _create_temp_files(self, root_path, num_files_per_extension,
                           file_extensions):
        """Create 'num_files_per_extension' temporary files
        per each of the given extensions.
        """
        files = set()
        for ext in file_extensions:
            for fileno in range(1, num_files_per_extension + 1):
                prefix = str(fileno)
                suffix = os.extsep + ext if ext else ''
                _, path = tempfile.mkstemp(prefix=prefix, suffix=suffix,
                                           dir=root_path)
                files.add(path)

        return files



#    Copyright 2014 Mirantis Inc.
#    All Rights Reserved.
#    Copyright 2015 Tesora Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from trove.guestagent.common import operating_system
from trove.guestagent.datastore.experimental.cassandra import service
from trove.guestagent.strategies.restore import base

LOG = logging.getLogger(__name__)


class NodetoolSnapshot(base.RestoreRunner):
    """Implementation of restore using the Nodetool (http://goo.gl/QtXVsM)
    utility.
    """

    __strategy_name__ = 'nodetoolsnapshot'

    def __init__(self, storage, **kwargs):
        self._app = service.CassandraApp()
        kwargs.update({'restore_location': self._app.cassandra_data_dir})
        super(NodetoolSnapshot, self).__init__(storage, **kwargs)

    def pre_restore(self):
        """Prepare the data directory for restored files.
        The directory itself is not included in the backup archive
        (i.e. the archive is rooted inside the data directory).
        This is to make sure we can always restore an old backup
        even if the standard guest agent data directory changes.
        """

        LOG.debug('Initializing a data directory.')
        operating_system.create_directory(
            self.restore_location,
            user=self._app.cassandra_owner, group=self._app.cassandra_owner,
            force=True, as_root=True)

    def post_restore(self):
        """Updated ownership on the restored files.
        """

        LOG.debug('Updating ownership of the restored files.')
        operating_system.chown(
            self.restore_location,
            self._app.cassandra_owner, self._app.cassandra_owner,
            recursive=True, force=True, as_root=True)

    @property
    def base_restore_cmd(self):
        """Command to extract a backup archive into a given location.
        Attempt to preserve access modifiers on the archived files.
        """

        return 'sudo tar -xpPf - -C "%(restore_location)s"'



# Copyright [2015] Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jsonschema

from mock import Mock, patch, MagicMock, PropertyMock
from testtools.matchers import Is, Equals

from trove.common import clients
from trove.common import exception
from trove.datastore import models as datastore_models
from trove.extensions.mgmt.datastores.service import DatastoreVersionController
from trove.tests.unittests import trove_testtools


class TestDatastoreVersionController(trove_testtools.TestCase):
    def setUp(self):
        super(TestDatastoreVersionController, self).setUp()
        self.controller = DatastoreVersionController()

        self.version = {
            "version": {
                "datastore_name": "test_dsx",
                "name": "test_vr1",
                "datastore_manager": "mysql",
                "image": "154b350d-4d86-4214-9067-9c54b230c0da",
                "packages": ["mysql-server-5.7"],
                "active": True,
                "default": False
            }
        }

        self.tenant_id = Mock()
        context = trove_testtools.TroveTestContext(self)
        self.req = Mock()
        self.req.environ = Mock()
        self.req.environ.__getitem__ = Mock(return_value=context)

    def test_get_schema_create(self):
        schema = self.controller.get_schema('create', self.version)
        self.assertIsNotNone(schema)
        self.assertIn('version', schema['properties'])

    def test_validate_create(self):
        body = self.version
        schema = self.controller.get_schema('create', body)
        validator = jsonschema.Draft4Validator(schema)
        self.assertTrue(validator.is_valid(body))

    def test_validate_create_blankname(self):
        body = self.version
        body['version']['name'] = "     "
        schema = self.controller.get_schema('create', body)
        validator = jsonschema.Draft4Validator(schema)
        self.assertFalse(validator.is_valid(body))
        errors = sorted(validator.iter_errors(body), key=lambda e: e.path)
        self.assertThat(len(errors), Is(1))
        self.assertThat(errors[0].message,
                        Equals("'     ' does not match '^.*[0-9a-zA-Z]+.*$'"))

    def test_validate_create_blank_datastore(self):
        body = self.version
        body['version']['datastore_name'] = ""
        schema = self.controller.get_schema('create', body)
        validator = jsonschema.Draft4Validator(schema)
        self.assertFalse(validator.is_valid(body))
        errors = sorted(validator.iter_errors(body), key=lambda e: e.path)
        error_messages = [error.message for error in errors]
        self.assertThat(len(errors), Is(2))
        self.assertIn("'' is too short", error_messages)
        self.assertIn("'' does not match '^.*[0-9a-zA-Z]+.*$'", error_messages)

    @patch.object(clients, 'create_glance_client')
    @patch.object(datastore_models.Datastore, 'load')
    @patch.object(datastore_models.DatastoreVersion, 'load',
                  side_effect=exception.DatastoreVersionNotFound)
    @patch.object(datastore_models, 'update_datastore_version')
    def test_create_datastore_versions(self, mock_ds_version_create,
                                       mock_ds_version_load,
                                       mock_ds_load, mock_glance_client):
        body = self.version
        mock_ds_load.return_value.name = 'test_dsx'

        self.controller.create(self.req, body, self.tenant_id)
        mock_ds_version_create.assert_called_with(
            'test_dsx', 'test_vr1', 'mysql',
            '154b350d-4d86-4214-9067-9c54b230c0da',
            'mysql-server-5.7', True)

    @patch.object(datastore_models.DatastoreVersion, 'load_by_uuid')
    def test_show_ds_version(self, mock_ds_version_load):
        id = Mock()

        self.controller.show(self.req, self.tenant_id, id)
        mock_ds_version_load.assert_called_with(id)

    @patch('trove.configuration.models.DBConfiguration.find_all')
    @patch('trove.backup.models.DBBackup.find_all')
    @patch('trove.instance.models.DBInstance.find_all')
    @patch.object(datastore_models.Datastore, 'load')
    @patch.object(datastore_models.DatastoreVersion, 'load_by_uuid')
    def test_delete_ds_version(self, mock_ds_version_load, mock_ds_load,
                               mock_instance_find, mock_backup_find,
                               mock_config_find):
        ds_version_id = Mock()
        ds_version = Mock()
        mock_ds_version_load.return_value = ds_version
        self.controller.delete(self.req, self.tenant_id, ds_version_id)
        ds_version.delete.assert_called_with()

    @patch('trove.instance.models.DBInstance.find_all')
    def test_delete_ds_version_instance_in_use(self, mock_instance_find):
        mock_instance_find.return_value.all.return_value = [Mock()]

        self.assertRaises(
            exception.DatastoreVersionsInUse,
            self.controller.delete,
            self.req, self.tenant_id, 'fake_version_id'
        )

    @patch('trove.backup.models.DBBackup.find_all')
    @patch('trove.instance.models.DBInstance.find_all')
    def test_delete_ds_version_backup_in_use(self, mock_instance_find,
                                             mock_backup_find):
        mock_backup_find.return_value.all.return_value = [Mock()]

        self.assertRaises(
            exception.DatastoreVersionsInUse,
            self.controller.delete,
            self.req, self.tenant_id, 'fake_version_id'
        )

    @patch('trove.configuration.models.DBConfiguration.find_all')
    @patch('trove.backup.models.DBBackup.find_all')
    @patch('trove.instance.models.DBInstance.find_all')
    def test_delete_ds_version_config_in_use(self, mock_instance_find,
                                             mock_backup_find,
                                             mock_config_find):
        mock_config_find.return_value.all.return_value = [Mock()]

        self.assertRaises(
            exception.DatastoreVersionsInUse,
            self.controller.delete,
            self.req, self.tenant_id, 'fake_version_id'
        )

    @patch.object(datastore_models.DatastoreVersion, 'load_by_uuid')
    @patch.object(datastore_models.DatastoreVersions, 'load_all')
    def test_index_ds_version(self, mock_ds_version_load_all,
                              mock_ds_version_load_by_uuid):
        mock_id = Mock()
        mock_ds_version = Mock()
        mock_ds_version.id = mock_id
        mock_ds_version_load_all.return_value = [mock_ds_version]

        self.controller.index(self.req, self.tenant_id)
        mock_ds_version_load_all.assert_called_with(only_active=False)
        mock_ds_version_load_by_uuid.assert_called_with(mock_id)

    @patch.object(clients, 'create_glance_client')
    @patch.object(datastore_models.DatastoreVersion, 'load_by_uuid')
    @patch.object(datastore_models, 'update_datastore_version')
    def test_edit_datastore_versions(self, mock_ds_version_update,
                                     mock_ds_version_load,
                                     mock_glance_client):
        body = {'image': '21c8805a-a800-4bca-a192-3a5a2519044d'}

        mock_ds_version = MagicMock()
        type(mock_ds_version).datastore_name = PropertyMock(
            return_value=self.version['version']['datastore_name'])
        type(mock_ds_version).name = PropertyMock(
            return_value=self.version['version']['name'])
        type(mock_ds_version).image_id = PropertyMock(
            return_value=self.version['version']['image'])
        type(mock_ds_version).packages = PropertyMock(
            return_value=self.version['version']['packages'])
        type(mock_ds_version).active = PropertyMock(
            return_value=self.version['version']['active'])
        type(mock_ds_version).manager = PropertyMock(
            return_value=self.version['version']['datastore_manager'])
        mock_ds_version_load.return_value = mock_ds_version

        self.controller.edit(self.req, body, self.tenant_id, Mock())
        mock_ds_version_update.assert_called_with(
            'test_dsx', 'test_vr1', 'mysql',
            '21c8805a-a800-4bca-a192-3a5a2519044d',
            'mysql-server-5.7', True)



# Copyright (c) 2014 eBay Software Foundation
# Copyright 2015 Hewlett-Packard Development Company, L.P. and Tesora, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging

from trove.common import cfg
from trove.common import exception
from trove.common.i18n import _
from trove.common import utils
from trove.guestagent.common import operating_system
from trove.guestagent.datastore.experimental.mongodb import (
    service as mongo_service)
from trove.guestagent.datastore.experimental.mongodb import (
    system as mongo_system)
from trove.guestagent.strategies.backup import base

CONF = cfg.CONF

LOG = logging.getLogger(__name__)
MONGODB_DBPATH = CONF.mongodb.mount_point
MONGO_DUMP_DIR = MONGODB_DBPATH + "/dump"
LARGE_TIMEOUT = 1200


class MongoDump(base.BackupRunner):
    """Implementation of Backup Strategy for MongoDump."""
    __strategy_name__ = 'mongodump'

    def __init__(self, *args, **kwargs):
        self.app = mongo_service.MongoDBApp()
        super(MongoDump, self).__init__(*args, **kwargs)

    def _run_pre_backup(self):
        """Create archival contents in dump dir"""
        try:
            est_dump_size = self.estimate_dump_size()
            avail = operating_system.get_bytes_free_on_fs(MONGODB_DBPATH)
            if est_dump_size > avail:
                self.cleanup()
                # TODO(atomic77) Though we can fully recover from this error
                # BackupRunner will leave the trove instance in a BACKUP state
                raise OSError(_("Need more free space to run mongodump, "
                                "estimated %(est_dump_size)s"
                                " and found %(avail)s bytes free ") %
                              {'est_dump_size': est_dump_size,
                               'avail': avail})

            operating_system.create_directory(MONGO_DUMP_DIR, as_root=True)
            operating_system.chown(MONGO_DUMP_DIR, mongo_system.MONGO_USER,
                                   mongo_system.MONGO_USER, as_root=True)

            # high timeout here since mongodump can take a long time
            utils.execute_with_timeout(
                'mongodump', '--out', MONGO_DUMP_DIR,
                *(self.app.admin_cmd_auth_params()),
                run_as_root=True, root_helper='sudo',
                timeout=LARGE_TIMEOUT
            )
        except exception.ProcessExecutionError:
            LOG.debug("Caught exception when creating the dump")
            self.cleanup()
            raise

    @property
    def cmd(self):
        """Tars and streams the dump dir contents to
        the stdout
        """
        cmd = 'sudo tar cPf - ' + MONGO_DUMP_DIR
        return cmd + self.zip_cmd + self.encrypt_cmd

    def cleanup(self):
        operating_system.remove(MONGO_DUMP_DIR, force=True, as_root=True)

    def _run_post_backup(self):
        self.cleanup()

    def estimate_dump_size(self):
        """
        Estimate the space that the mongodump will take based on the output of
        db.stats().dataSize. This seems to be conservative, as the actual bson
        output in many cases is a fair bit smaller.
        """
        dbs = self.app.list_all_dbs()
        # mongodump does not dump the content of the local database
        dbs.remove('local')
        dbstats = dict([(d, 0) for d in dbs])
        for d in dbstats:
            dbstats[d] = self.app.db_data_size(d)

        LOG.debug("Estimated size for databases: " + str(dbstats))
        return sum(dbstats.values())



# Copyright 2015 Hewlett-Packard Development Company, L.P. and Tesora, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from oslo_log import log as logging

from trove.guestagent.common import operating_system
from trove.guestagent.common.operating_system import FileMode
from trove.guestagent.datastore.experimental.redis import service
from trove.guestagent.datastore.experimental.redis import system
from trove.guestagent.strategies.restore import base

LOG = logging.getLogger(__name__)


class RedisBackup(base.RestoreRunner):
    """Implementation of Restore Strategy for Redis."""
    __strategy_name__ = 'redisbackup'

    CONF_LABEL_AOF_TEMP_OFF = 'restore_aof_temp_off'
    INFO_PERSISTENCE_SECTION = 'persistence'

    def __init__(self, storage, **kwargs):
        self.app = service.RedisApp()
        self.restore_location = self.app.get_persistence_filepath()
        self.base_restore_cmd = 'tee %s' % self.restore_location
        self.aof_set = self.app.is_appendonly_enabled()
        self.aof_off_cfg = {'appendonly': 'no'}
        kwargs.update({'restore_location': self.restore_location})
        super(RedisBackup, self).__init__(storage, **kwargs)

    def pre_restore(self):
        self.app.stop_db()
        LOG.info("Removing old persistence file: %s.",
                 self.restore_location)
        operating_system.remove(self.restore_location, force=True,
                                as_root=True)
        dir = os.path.dirname(self.restore_location)
        operating_system.create_directory(dir, as_root=True)
        operating_system.chmod(dir, FileMode.SET_FULL, as_root=True)
        # IF AOF is set, we need to turn it off temporarily
        if self.aof_set:
            self.app.configuration_manager.apply_system_override(
                self.aof_off_cfg, change_id=self.CONF_LABEL_AOF_TEMP_OFF)

    def post_restore(self):
        operating_system.chown(self.restore_location,
                               system.REDIS_OWNER, system.REDIS_OWNER,
                               as_root=True)
        self.app.start_db()

        # IF AOF was set, we need to put back the original file
        if self.aof_set:
            self.app.admin.wait_until('loading', 0,
                                      section=self.INFO_PERSISTENCE_SECTION)
            self.app.admin.execute('BGREWRITEAOF')
            self.app.admin.wait_until('aof_rewrite_in_progress', 0,
                                      section=self.INFO_PERSISTENCE_SECTION)
            self.app.stop_db()
            self.app.configuration_manager.remove_system_override(
                change_id=self.CONF_LABEL_AOF_TEMP_OFF)
            self.app.start_db()



#    Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import random
import string

from mock import ANY
from mock import call
from mock import DEFAULT
from mock import MagicMock
from mock import Mock
from mock import NonCallableMagicMock
from mock import patch
from oslo_utils import netutils
from testtools import ExpectedException

from trove.common.db.cassandra import models
from trove.common import exception
from trove.common.instance import ServiceStatuses
from trove.guestagent import backup
from trove.guestagent.common.configuration import ImportOverrideStrategy
from trove.guestagent.common import operating_system
from trove.guestagent.datastore.experimental.cassandra import (
    manager as cass_manager)
from trove.guestagent.datastore.experimental.cassandra import (
    service as cass_service)
from trove.guestagent import pkg
from trove.guestagent import volume
from trove.tests.unittests.guestagent.test_datastore_manager import \
    DatastoreManagerTest
from trove.tests.unittests import trove_testtools


class GuestAgentCassandraDBManagerTest(DatastoreManagerTest):

    __MOUNT_POINT = '/var/lib/cassandra'

    __N_GAK = '_get_available_keyspaces'
    __N_GLU = '_get_listed_users'
    __N_BU = '_build_user'
    __N_RU = '_rename_user'
    __N_AUP = '_alter_user_password'
    __N_CAU = 'trove.common.db.cassandra.models.CassandraUser'
    __N_CU = '_create_user'
    __N_GFA = '_grant_full_access_on_keyspace'
    __N_DU = '_drop_user'

    __ACCESS_MODIFIERS = ('ALTER', 'CREATE', 'DROP', 'MODIFY', 'SELECT')
    __CREATE_DB_FORMAT = (
        "CREATE KEYSPACE \"{}\" WITH REPLICATION = "
        "{{ 'class' : 'SimpleStrategy', 'replication_factor' : 1 }};"
    )
    __DROP_DB_FORMAT = "DROP KEYSPACE \"{}\";"
    __CREATE_USR_FORMAT = "CREATE USER '{}' WITH PASSWORD %s NOSUPERUSER;"
    __ALTER_USR_FORMAT = "ALTER USER '{}' WITH PASSWORD %s;"
    __DROP_USR_FORMAT = "DROP USER '{}';"
    __GRANT_FORMAT = "GRANT {} ON KEYSPACE \"{}\" TO '{}';"
    __REVOKE_FORMAT = "REVOKE ALL PERMISSIONS ON KEYSPACE \"{}\" FROM '{}';"
    __LIST_PERMISSIONS_FORMAT = "LIST ALL PERMISSIONS NORECURSIVE;"
    __LIST_PERMISSIONS_OF_FORMAT = "LIST ALL PERMISSIONS OF '{}' NORECURSIVE;"
    __LIST_DB_FORMAT = "SELECT * FROM system.schema_keyspaces;"
    __LIST_USR_FORMAT = "LIST USERS;"

    @patch.object(ImportOverrideStrategy, '_initialize_import_directory')
    @patch('trove.guestagent.datastore.experimental.cassandra.service.LOG')
    def setUp(self, *args, **kwargs):
        super(GuestAgentCassandraDBManagerTest, self).setUp('cassandra')

        conn_patcher = patch.multiple(cass_service.CassandraConnection,
                                      _connect=DEFAULT,
                                      is_active=Mock(return_value=True))
        self.addCleanup(conn_patcher.stop)
        conn_patcher.start()

        self.real_status = cass_service.CassandraAppStatus.set_status

        class FakeInstanceServiceStatus(object):
            status = ServiceStatuses.NEW

            def save(self):
                pass

        cass_service.CassandraAppStatus.set_status = MagicMock(
            return_value=FakeInstanceServiceStatus())
        self.context = trove_testtools.TroveTestContext(self)
        self.manager = cass_manager.Manager()
        self.manager._app = cass_service.CassandraApp()
        self.manager._admin = cass_service.CassandraAdmin(
            models.CassandraUser('Test'))
        self.admin = self.manager._admin
        self.admin._CassandraAdmin__client = MagicMock()
        self.conn = self.admin._CassandraAdmin__client
        self.pkg = cass_service.packager
        self.origin_os_path_exists = os.path.exists
        self.origin_format = volume.VolumeDevice.format
        self.origin_migrate_data = volume.VolumeDevice.migrate_data
        self.origin_mount = volume.VolumeDevice.mount
        self.origin_mount_points = volume.VolumeDevice.mount_points
        self.origin_stop_db = cass_service.CassandraApp.stop_db
        self.origin_start_db = cass_service.CassandraApp.start_db
        self.origin_install_db = cass_service.CassandraApp._install_db
        self.original_get_ip = netutils.get_my_ipv4
        self.orig_make_host_reachable = (
            cass_service.CassandraApp.apply_initial_guestagent_configuration)

    def tearDown(self):
        super(GuestAgentCassandraDBManagerTest, self).tearDown()
        cass_service.packager = self.pkg
        os.path.exists = self.origin_os_path_exists
        volume.VolumeDevice.format = self.origin_format
        volume.VolumeDevice.migrate_data = self.origin_migrate_data
        volume.VolumeDevice.mount = self.origin_mount
        volume.VolumeDevice.mount_points = self.origin_mount_points
        cass_service.CassandraApp.stop_db = self.origin_stop_db
        cass_service.CassandraApp.start_db = self.origin_start_db
        cass_service.CassandraApp._install_db = self.origin_install_db
        netutils.get_my_ipv4 = self.original_get_ip
        cass_service.CassandraApp.apply_initial_guestagent_configuration = (
            self.orig_make_host_reachable)
        cass_service.CassandraAppStatus.set_status = self.real_status

    def test_update_status(self):
        mock_status = MagicMock()
        mock_status.is_installed = True
        mock_status._is_restarting = False
        self.manager._app.status = mock_status

        self.manager.update_status(self.context)

        self.assertTrue(mock_status.set_status.called)

    def test_prepare_pkg(self):
        self._prepare_dynamic(['cassandra'])

    def test_prepare_no_pkg(self):
        self._prepare_dynamic([])

    def test_prepare_db_not_installed(self):
        self._prepare_dynamic([], is_db_installed=False)

    def test_prepare_db_not_installed_no_package(self):
        self._prepare_dynamic([],
                              is_db_installed=True)

    @patch.object(backup, 'restore')
    def test_prepare_db_restore(self, restore):
        backup_info = {'id': 'backup_id',
                       'instance_id': 'fake-instance-id',
                       'location': 'fake-location',
                       'type': 'InnoBackupEx',
                       'checksum': 'fake-checksum'}

        self._prepare_dynamic(['cassandra'], is_db_installed=False,
                              backup_info=backup_info)
        restore.assert_called_once_with(
            self.context, backup_info, self.__MOUNT_POINT)

    @patch.multiple(operating_system, enable_service_on_boot=DEFAULT,
                    disable_service_on_boot=DEFAULT)
    @patch('trove.guestagent.datastore.experimental.cassandra.service.LOG')
    def test_superuser_password_reset(
            self, _, enable_service_on_boot, disable_service_on_boot):
        fake_status = MagicMock()
        fake_status.is_running = False

        test_app = cass_service.CassandraApp()
        test_app.status = fake_status
        with patch.multiple(
                test_app,
                start_db=DEFAULT,
                stop_db=DEFAULT,
                restart=DEFAULT,
                _CassandraApp__disable_remote_access=DEFAULT,
                _CassandraApp__enable_remote_access=DEFAULT,
                _CassandraApp__disable_authentication=DEFAULT,
                _CassandraApp__enable_authentication=DEFAULT,
                _CassandraApp__reset_user_password_to_default=DEFAULT,
                secure=DEFAULT) as calls:

            test_app._reset_admin_password()

            disable_service_on_boot.assert_called_once_with(
                test_app.service_candidates)
            calls[
                '_CassandraApp__disable_remote_access'
            ].assert_called_once_with()
            calls[
                '_CassandraApp__disable_authentication'
            ].assert_called_once_with()
            calls['start_db'].assert_called_once_with(update_db=False,
                                                      enable_on_boot=False),
            calls[
                '_CassandraApp__enable_authentication'
            ].assert_called_once_with()

            pw_reset_mock = calls[
                '_CassandraApp__reset_user_password_to_default'
            ]
            pw_reset_mock.assert_called_once_with(test_app._ADMIN_USER)
            calls['secure'].assert_called_once_with(
                update_user=pw_reset_mock.return_value)
            calls['restart'].assert_called_once_with()
            calls['stop_db'].assert_called_once_with()
            calls[
                '_CassandraApp__enable_remote_access'
            ].assert_called_once_with()
            enable_service_on_boot.assert_called_once_with(
                test_app.service_candidates)

    @patch('trove.guestagent.datastore.experimental.cassandra.service.LOG')
    def test_change_cluster_name(self, _):
        fake_status = MagicMock()
        fake_status.is_running = True

        test_app = cass_service.CassandraApp()
        test_app.status = fake_status
        with patch.multiple(
                test_app,
                start_db=DEFAULT,
                stop_db=DEFAULT,
                restart=DEFAULT,
                _update_cluster_name_property=DEFAULT,
                _CassandraApp__reset_cluster_name=DEFAULT) as calls:

            sample_name = NonCallableMagicMock()
            test_app.change_cluster_name(sample_name)
            calls['_CassandraApp__reset_cluster_name'].assert_called_once_with(
                sample_name)
            calls['_update_cluster_name_property'].assert_called_once_with(
                sample_name)
            calls['restart'].assert_called_once_with()

    @patch.object(cass_service, 'CONF', DEFAULT)
    @patch('trove.guestagent.datastore.experimental.cassandra.service.LOG')
    def test_apply_post_restore_updates(self, _, conf_mock):
        fake_status = MagicMock()
        fake_status.is_running = False

        test_app = cass_service.CassandraApp()
        test_app.status = fake_status
        with patch.multiple(
                test_app,
                start_db=DEFAULT,
                stop_db=DEFAULT,
                _update_cluster_name_property=DEFAULT,
                _reset_admin_password=DEFAULT,
                change_cluster_name=DEFAULT) as calls:
            backup_info = {'instance_id': 'old_id'}
            conf_mock.guest_id = 'new_id'
            test_app._apply_post_restore_updates(backup_info)
            calls['_update_cluster_name_property'].assert_called_once_with(
                'old_id')
            calls['_reset_admin_password'].assert_called_once_with()
            calls['start_db'].assert_called_once_with(update_db=False)
            calls['change_cluster_name'].assert_called_once_with('new_id')
            calls['stop_db'].assert_called_once_with()

    def _prepare_dynamic(self, packages,
                         config_content='MockContent', device_path='/dev/vdb',
                         is_db_installed=True, backup_info=None,
                         is_root_enabled=False,
                         overrides=None):

        mock_status = MagicMock()
        mock_app = MagicMock()
        mock_app.status = mock_status
        self.manager._app = mock_app

        mock_status.begin_install = MagicMock(return_value=None)
        mock_app.install_if_needed = MagicMock(return_value=None)
        mock_app.init_storage_structure = MagicMock(return_value=None)
        mock_app.write_config = MagicMock(return_value=None)
        mock_app.apply_initial_guestagent_configuration = MagicMock(
            return_value=None)
        mock_app.restart = MagicMock(return_value=None)
        mock_app.start_db = MagicMock(return_value=None)
        mock_app.stop_db = MagicMock(return_value=None)
        mock_app._remove_system_tables = MagicMock(return_value=None)
        os.path.exists = MagicMock(return_value=True)
        volume.VolumeDevice.format = MagicMock(return_value=None)
        volume.VolumeDevice.migrate_data = MagicMock(return_value=None)
        volume.VolumeDevice.mount = MagicMock(return_value=None)
        volume.VolumeDevice.mount_points = MagicMock(return_value=[])

        with patch.object(pkg.Package, 'pkg_is_installed',
                          return_value=is_db_installed):
            # invocation
            self.manager.prepare(context=self.context, packages=packages,
                                 config_contents=config_content,
                                 databases=None,
                                 memory_mb='2048', users=None,
                                 device_path=device_path,
                                 mount_point=self.__MOUNT_POINT,
                                 backup_info=backup_info,
                                 overrides=None,
                                 cluster_config=None)

        # verification/assertion
        mock_status.begin_install.assert_any_call()
        mock_app.install_if_needed.assert_any_call(packages)
        mock_app._remove_system_tables.assert_any_call()
        mock_app.init_storage_structure.assert_any_call('/var/lib/cassandra')
        mock_app.apply_initial_guestagent_configuration.assert_any_call(
            cluster_name=None)
        mock_app.start_db.assert_any_call(update_db=False)
        mock_app.stop_db.assert_any_call()
        if backup_info:
            mock_app._apply_post_restore_updates.assert_called_once_with(
                backup_info)

    def test_keyspace_validation(self):
        valid_name = self._get_random_name(32)
        db = models.CassandraSchema(valid_name)
        self.assertEqual(valid_name, db.name)
        with ExpectedException(ValueError):
            models.CassandraSchema(self._get_random_name(33))

    def test_user_validation(self):
        valid_name = self._get_random_name(65535)
        usr = models.CassandraUser(valid_name, 'password')
        self.assertEqual(valid_name, usr.name)
        self.assertEqual('password', usr.password)
        with ExpectedException(ValueError):
            models.CassandraUser(self._get_random_name(65536))

    @classmethod
    def _serialize_collection(cls, *collection):
        return [item.serialize() for item in collection]

    @classmethod
    def _get_random_name(cls, size,
                         chars=string.ascii_letters + string.digits):
        return ''.join(random.choice(chars) for _ in range(size))

    def test_create_database(self):
        db1 = models.CassandraSchema('db1')
        db2 = models.CassandraSchema('db2')
        db3 = models.CassandraSchema(self._get_random_name(32))

        self.manager.create_database(self.context,
                                     self._serialize_collection(db1, db2, db3))
        self.conn.execute.assert_has_calls([
            call(self.__CREATE_DB_FORMAT, (db1.name,)),
            call(self.__CREATE_DB_FORMAT, (db2.name,)),
            call(self.__CREATE_DB_FORMAT, (db3.name,))
        ])

    def test_delete_database(self):
        db = models.CassandraSchema(self._get_random_name(32))
        self.manager.delete_database(self.context, db.serialize())
        self.conn.execute.assert_called_once_with(
            self.__DROP_DB_FORMAT, (db.name,))

    def test_create_user(self):
        usr1 = models.CassandraUser('usr1')
        usr2 = models.CassandraUser('usr2', '')
        usr3 = models.CassandraUser(self._get_random_name(1025), 'password')

        self.manager.create_user(self.context,
                                 self._serialize_collection(usr1, usr2, usr3))
        self.conn.execute.assert_has_calls([
            call(self.__CREATE_USR_FORMAT, (usr1.name,), (usr1.password,)),
            call(self.__CREATE_USR_FORMAT, (usr2.name,), (usr2.password,)),
            call(self.__CREATE_USR_FORMAT, (usr3.name,), (usr3.password,))
        ])

    def test_delete_user(self):
        usr = models.CassandraUser(self._get_random_name(1025), 'password')
        self.manager.delete_user(self.context, usr.serialize())
        self.conn.execute.assert_called_once_with(
            self.__DROP_USR_FORMAT, (usr.name,))

    def test_change_passwords(self):
        usr1 = models.CassandraUser('usr1')
        usr2 = models.CassandraUser('usr2', '')
        usr3 = models.CassandraUser(self._get_random_name(1025), 'password')

        self.manager.change_passwords(self.context, self._serialize_collection(
            usr1, usr2, usr3))
        self.conn.execute.assert_has_calls([
            call(self.__ALTER_USR_FORMAT, (usr1.name,), (usr1.password,)),
            call(self.__ALTER_USR_FORMAT, (usr2.name,), (usr2.password,)),
            call(self.__ALTER_USR_FORMAT, (usr3.name,), (usr3.password,))
        ])

    def test_alter_user_password(self):
        usr1 = models.CassandraUser('usr1')
        usr2 = models.CassandraUser('usr2', '')
        usr3 = models.CassandraUser(self._get_random_name(1025), 'password')

        self.admin.alter_user_password(usr1)
        self.admin.alter_user_password(usr2)
        self.admin.alter_user_password(usr3)
        self.conn.execute.assert_has_calls([
            call(self.__ALTER_USR_FORMAT, (usr1.name,), (usr1.password,)),
            call(self.__ALTER_USR_FORMAT, (usr2.name,), (usr2.password,)),
            call(self.__ALTER_USR_FORMAT, (usr3.name,), (usr3.password,))
        ])

    def test_grant_access(self):
        usr1 = models.CassandraUser('usr1')
        usr2 = models.CassandraUser('usr1', 'password')
        db1 = models.CassandraSchema('db1')
        db2 = models.CassandraSchema('db2')
        db3 = models.CassandraSchema('db3')

        self.manager.grant_access(self.context, usr1.name, None, [db1.name,
                                                                  db2.name])
        self.manager.grant_access(self.context, usr2.name, None, [db3.name])

        expected = []
        for modifier in self.__ACCESS_MODIFIERS:
            expected.append(call(self.__GRANT_FORMAT,
                                 (modifier, db1.name, usr1.name)))
            expected.append(call(self.__GRANT_FORMAT,
                                 (modifier, db3.name, usr2.name)))

        self.conn.execute.assert_has_calls(
            expected,
            any_order=True)

    def test_revoke_access(self):
        usr1 = models.CassandraUser('usr1')
        usr2 = models.CassandraUser('usr1', 'password')
        db1 = models.CassandraSchema('db1')
        db2 = models.CassandraSchema('db2')

        self.manager.revoke_access(self.context, usr1.name, None, db1.name)
        self.manager.revoke_access(self.context, usr2.name, None, db2.name)
        self.conn.execute.assert_has_calls([
            call(self.__REVOKE_FORMAT, (db1.name, usr1.name)),
            call(self.__REVOKE_FORMAT, (db2.name, usr2.name))
        ])

    def test_get_available_keyspaces(self):
        self.manager.list_databases(self.context)
        self.conn.execute.assert_called_once_with(
            self.__LIST_DB_FORMAT)

    def test_list_databases(self):
        db1 = models.CassandraSchema('db1')
        db2 = models.CassandraSchema('db2')
        db3 = models.CassandraSchema(self._get_random_name(32))

        with patch.object(self.admin, self.__N_GAK, return_value={db1, db2,
                                                                  db3}):
            found = self.manager.list_databases(self.context)
            self.assertEqual(2, len(found))
            self.assertEqual(3, len(found[0]))
            self.assertIsNone(found[1])
            self.assertIn(db1.serialize(), found[0])
            self.assertIn(db2.serialize(), found[0])
            self.assertIn(db3.serialize(), found[0])

        with patch.object(self.admin, self.__N_GAK, return_value=set()):
            found = self.manager.list_databases(self.context)
            self.assertEqual(([], None), found)

    def test_get_acl(self):
        r0 = NonCallableMagicMock(username='user1', resource='<all keyspaces>',
                                  permission='SELECT')
        r1 = NonCallableMagicMock(username='user2', resource='<keyspace ks1>',
                                  permission='SELECT')
        r2 = NonCallableMagicMock(username='user2', resource='<keyspace ks2>',
                                  permission='SELECT')
        r3 = NonCallableMagicMock(username='user2', resource='<keyspace ks2>',
                                  permission='ALTER')
        r4 = NonCallableMagicMock(username='user3', resource='<table ks2.t1>',
                                  permission='SELECT')
        r5 = NonCallableMagicMock(username='user3', resource='',
                                  permission='ALTER')
        r6 = NonCallableMagicMock(username='user3', resource='<keyspace ks2>',
                                  permission='')
        r7 = NonCallableMagicMock(username='user3', resource='',
                                  permission='')
        r8 = NonCallableMagicMock(username='user3', resource='<keyspace ks1>',
                                  permission='DELETE')
        r9 = NonCallableMagicMock(username='user4', resource='<all keyspaces>',
                                  permission='UPDATE')
        r10 = NonCallableMagicMock(username='user4', resource='<keyspace ks1>',
                                   permission='DELETE')

        available_ks = {models.CassandraSchema('ks1'),
                        models.CassandraSchema('ks2'),
                        models.CassandraSchema('ks3')}

        mock_result_set = [r0, r1, r2, r3, r4, r5, r6, r7, r8, r9, r9, r9, r10]
        execute_mock = MagicMock(return_value=mock_result_set)
        mock_client = MagicMock(execute=execute_mock)

        with patch.object(self.admin,
                          self.__N_GAK, return_value=available_ks) as gak_mock:
            acl = self.admin._get_acl(mock_client)
            execute_mock.assert_called_once_with(
                self.__LIST_PERMISSIONS_FORMAT)
            gak_mock.assert_called_once_with(mock_client)

            self.assertEqual({'user1': {'ks1': {'SELECT'},
                                        'ks2': {'SELECT'},
                                        'ks3': {'SELECT'}},
                              'user2': {'ks1': {'SELECT'},
                                        'ks2': {'SELECT', 'ALTER'}},
                              'user3': {'ks1': {'DELETE'}},
                              'user4': {'ks1': {'UPDATE', 'DELETE'},
                                        'ks2': {'UPDATE'},
                                        'ks3': {'UPDATE'}}
                              },
                             acl)

        mock_result_set = [r1, r2, r3]
        execute_mock = MagicMock(return_value=mock_result_set)
        mock_client = MagicMock(execute=execute_mock)

        with patch.object(self.admin,
                          self.__N_GAK, return_value=available_ks) as gak_mock:
            acl = self.admin._get_acl(mock_client, username='user2')
            execute_mock.assert_called_once_with(
                self.__LIST_PERMISSIONS_OF_FORMAT.format('user2'))
            gak_mock.assert_not_called()

            self.assertEqual({'user2': {'ks1': {'SELECT'},
                                        'ks2': {'SELECT', 'ALTER'}}}, acl)

        mock_result_set = []
        execute_mock = MagicMock(return_value=mock_result_set)
        mock_client = MagicMock(execute=execute_mock)

        with patch.object(self.admin,
                          self.__N_GAK, return_value=available_ks) as gak_mock:
            acl = self.admin._get_acl(mock_client, username='nonexisting')
            execute_mock.assert_called_once_with(
                self.__LIST_PERMISSIONS_OF_FORMAT.format('nonexisting'))
            gak_mock.assert_not_called()

            self.assertEqual({}, acl)

    def test_get_listed_users(self):
        usr1 = models.CassandraUser(self._get_random_name(1025))
        usr2 = models.CassandraUser(self._get_random_name(1025))
        usr3 = models.CassandraUser(self._get_random_name(1025))
        db1 = models.CassandraSchema('db1')
        db2 = models.CassandraSchema('db2')
        usr1.databases.append(db1.serialize())
        usr3.databases.append(db2.serialize())

        rv_1 = NonCallableMagicMock()
        rv_1.configure_mock(name=usr1.name, super=False)
        rv_2 = NonCallableMagicMock()
        rv_2.configure_mock(name=usr2.name, super=False)
        rv_3 = NonCallableMagicMock()
        rv_3.configure_mock(name=usr3.name, super=True)

        with patch.object(self.conn, 'execute', return_value=iter(
                [rv_1, rv_2, rv_3])):
            with patch.object(self.admin, '_get_acl',
                              return_value={usr1.name: {db1.name: {'SELECT'},
                                                        db2.name: {}},
                                            usr3.name: {db2.name: {'SELECT'}}}
                              ):
                usrs = self.manager.list_users(self.context)
                self.conn.execute.assert_has_calls([
                    call(self.__LIST_USR_FORMAT),
                ], any_order=True)
                self.assertIn(usr1.serialize(), usrs[0])
                self.assertIn(usr2.serialize(), usrs[0])
                self.assertIn(usr3.serialize(), usrs[0])

    def test_list_access(self):
        usr1 = models.CassandraUser('usr1')
        usr2 = models.CassandraUser('usr2')
        usr3 = models.CassandraUser(self._get_random_name(1025), 'password')
        db1 = models.CassandraSchema('db1').serialize()
        db2 = models.CassandraSchema('db2').serialize()
        usr2.databases.append(db1)
        usr3.databases.append(db1)
        usr3.databases.append(db2)

        with patch.object(self.admin, self.__N_GLU, return_value={usr1, usr2,
                                                                  usr3}):
            usr1_dbs = self.manager.list_access(self.context, usr1.name, None)
            usr2_dbs = self.manager.list_access(self.context, usr2.name, None)
            usr3_dbs = self.manager.list_access(self.context, usr3.name, None)
            self.assertEqual([], usr1_dbs)
            self.assertEqual([db1], usr2_dbs)
            self.assertEqual([db1, db2], usr3_dbs)

        with patch.object(self.admin, self.__N_GLU, return_value=set()):
            with ExpectedException(exception.UserNotFound):
                self.manager.list_access(self.context, usr3.name, None)

    def test_list_users(self):
        usr1 = models.CassandraUser('usr1')
        usr2 = models.CassandraUser('usr2')
        usr3 = models.CassandraUser(self._get_random_name(1025), 'password')

        with patch.object(self.admin, self.__N_GLU, return_value={usr1, usr2,
                                                                  usr3}):
            found = self.manager.list_users(self.context)
            self.assertEqual(2, len(found))
            self.assertEqual(3, len(found[0]))
            self.assertIsNone(found[1])
            self.assertIn(usr1.serialize(), found[0])
            self.assertIn(usr2.serialize(), found[0])
            self.assertIn(usr3.serialize(), found[0])

        with patch.object(self.admin, self.__N_GLU, return_value=set()):
            self.assertEqual(([], None), self.manager.list_users(self.context))

    def test_get_user(self):
        usr1 = models.CassandraUser('usr1')
        usr2 = models.CassandraUser('usr2')
        usr3 = models.CassandraUser(self._get_random_name(1025), 'password')

        with patch.object(self.admin, self.__N_GLU, return_value={usr1, usr2,
                                                                  usr3}):
            found = self.manager.get_user(self.context, usr2.name, None)
            self.assertEqual(usr2.serialize(), found)

        with patch.object(self.admin, self.__N_GLU, return_value=set()):
            self.assertIsNone(
                self.manager.get_user(self.context, usr2.name, None))

    @patch.object(cass_service.CassandraAdmin, '_deserialize_keyspace',
                  side_effect=lambda p1: p1)
    def test_rename_user(self, ks_deserializer):
        usr = models.CassandraUser('usr')
        db1 = models.CassandraSchema('db1').serialize()
        db2 = models.CassandraSchema('db2').serialize()
        usr.databases.append(db1)
        usr.databases.append(db2)

        new_user = models.CassandraUser('new_user')
        with patch(self.__N_CAU, return_value=new_user):
            with patch.object(self.admin, self.__N_BU, return_value=usr):
                with patch.object(self.admin, self.__N_CU) as create:
                    with patch.object(self.admin, self.__N_GFA) as grant:
                        with patch.object(self.admin, self.__N_DU) as drop:
                            usr_attrs = {'name': 'user', 'password': 'trove'}
                            self.manager.update_attributes(self.context,
                                                           usr.name, None,
                                                           usr_attrs)
                            create.assert_called_once_with(ANY, new_user)
                            grant.assert_has_calls([call(ANY, db1, ANY),
                                                    call(ANY, db2, ANY)])
                            drop.assert_called_once_with(ANY, usr)

    def test_update_attributes(self):
        usr = models.CassandraUser('usr', 'pwd')

        with patch.object(self.admin, self.__N_BU, return_value=usr):
            usr_attrs = {'name': usr.name, 'password': usr.password}
            with patch.object(self.admin, self.__N_RU) as rename:
                with patch.object(self.admin, self.__N_AUP) as alter:
                    self.manager.update_attributes(self.context, usr.name,
                                                   None, usr_attrs)
                    self.assertEqual(0, rename.call_count)
                    self.assertEqual(0, alter.call_count)

            usr_attrs = {'name': 'user', 'password': 'password'}
            with patch.object(self.admin, self.__N_RU) as rename:
                with patch.object(self.admin, self.__N_AUP) as alter:
                    self.manager.update_attributes(self.context, usr.name,
                                                   None, usr_attrs)
                    rename.assert_called_once_with(ANY, usr, usr_attrs['name'],
                                                   usr_attrs['password'])
                    self.assertEqual(0, alter.call_count)

            usr_attrs = {'name': 'user', 'password': usr.password}
            with patch.object(self.admin, self.__N_RU) as rename:
                with patch.object(self.admin, self.__N_AUP) as alter:
                    self.manager.update_attributes(self.context, usr.name,
                                                   None, usr_attrs)
                    rename.assert_called_once_with(ANY, usr, usr_attrs['name'],
                                                   usr_attrs['password'])
                    self.assertEqual(0, alter.call_count)

            usr_attrs = {'name': 'user'}
            with patch.object(self.admin, self.__N_RU) as rename:
                with patch.object(self.admin, self.__N_AUP) as alter:
                    with ExpectedException(
                            exception.UnprocessableEntity, "Updating username "
                            "requires specifying a password as well."):
                        self.manager.update_attributes(self.context, usr.name,
                                                       None, usr_attrs)
                        self.assertEqual(0, rename.call_count)
                        self.assertEqual(0, alter.call_count)

            usr_attrs = {'name': usr.name, 'password': 'password'}
            with patch.object(self.admin, self.__N_RU) as rename:
                with patch.object(self.admin, self.__N_AUP) as alter:
                    self.manager.update_attributes(self.context, usr.name,
                                                   None, usr_attrs)
                    alter.assert_called_once_with(ANY, usr)
                    self.assertEqual(0, rename.call_count)

            usr_attrs = {'password': usr.password}
            with patch.object(self.admin, self.__N_RU) as rename:
                with patch.object(self.admin, self.__N_AUP) as alter:
                    self.manager.update_attributes(self.context, usr.name,
                                                   None, usr_attrs)
                    self.assertEqual(0, rename.call_count)
                    self.assertEqual(0, alter.call_count)

            usr_attrs = {'password': 'trove'}
            with patch.object(self.admin, self.__N_RU) as rename:
                with patch.object(self.admin, self.__N_AUP) as alter:
                    self.manager.update_attributes(self.context, usr.name,
                                                   None, usr_attrs)
                    alter.assert_called_once_with(ANY, usr)
                    self.assertEqual(0, rename.call_count)

    def test_update_overrides(self):
        cfg_mgr_mock = MagicMock()
        self.manager._app.configuration_manager = cfg_mgr_mock
        overrides = NonCallableMagicMock()
        self.manager.update_overrides(Mock(), overrides)
        cfg_mgr_mock.apply_user_override.assert_called_once_with(overrides)
        cfg_mgr_mock.remove_user_override.assert_not_called()

    def test_remove_overrides(self):
        cfg_mgr_mock = MagicMock()
        self.manager._app.configuration_manager = cfg_mgr_mock
        self.manager.update_overrides(Mock(), {}, remove=True)
        cfg_mgr_mock.remove_user_override.assert_called_once_with()
        cfg_mgr_mock.apply_user_override.assert_not_called()

    def test_apply_overrides(self):
        self.assertIsNone(
            self.manager.apply_overrides(Mock(), NonCallableMagicMock()))

    @patch('trove.guestagent.datastore.experimental.cassandra.service.LOG')
    @patch.object(cass_service.CassandraApp, '_run_nodetool_command')
    def test_drain(self, command_runner_mock, _):
        self.manager._app.drain()
        command_runner_mock.assert_called_once_with('drain')

    @patch('trove.guestagent.datastore.experimental.cassandra.service.LOG')
    @patch.object(cass_service.CassandraApp, '_run_nodetool_command')
    def test_upgrade_sstables(self, command_runner_mock, _):
        self.manager._app.upgrade_sstables()
        command_runner_mock.assert_called_once_with('upgradesstables')

    @patch('trove.guestagent.datastore.experimental.cassandra.service.LOG')
    def test_enable_root(self, _):
        with patch.object(self.manager._app, 'is_root_enabled',
                          return_value=False):
            with patch.object(cass_service.CassandraAdmin,
                              '_create_superuser') as create_mock:
                self.manager.enable_root(self.context)
                create_mock.assert_called_once_with(ANY)

        with patch.object(self.manager._app, 'is_root_enabled',
                          return_value=True):
            with patch.object(cass_service.CassandraAdmin,
                              'alter_user_password') as alter_mock:
                self.manager.enable_root(self.context)
                alter_mock.assert_called_once_with(ANY)

    @patch('trove.guestagent.datastore.experimental.cassandra.service.LOG')
    def test_is_root_enabled(self, _):
        trove_admin = Mock()
        trove_admin.configure_mock(name=self.manager._app._ADMIN_USER)
        other_admin = Mock()
        other_admin.configure_mock(name='someuser')

        with patch.object(cass_service.CassandraAdmin,
                          'list_superusers', return_value=[]):
            self.assertFalse(self.manager.is_root_enabled(self.context))

        with patch.object(cass_service.CassandraAdmin,
                          'list_superusers', return_value=[trove_admin]):
            self.assertFalse(self.manager.is_root_enabled(self.context))

        with patch.object(cass_service.CassandraAdmin,
                          'list_superusers', return_value=[other_admin]):
            self.assertTrue(self.manager.is_root_enabled(self.context))

        with patch.object(cass_service.CassandraAdmin,
                          'list_superusers',
                          return_value=[trove_admin, other_admin]):
            self.assertTrue(self.manager.is_root_enabled(self.context))

    def test_guest_log_enable(self):
        self._assert_guest_log_enable(False, 'INFO')
        self._assert_guest_log_enable(True, 'OFF')

    def _assert_guest_log_enable(self, disable, expected_level):
        with patch.multiple(
                self.manager._app,
                logback_conf_manager=DEFAULT,
                _run_nodetool_command=DEFAULT
        ) as app_mocks:
            self.assertFalse(self.manager.guest_log_enable(
                Mock(), Mock(), disable))

            (app_mocks['logback_conf_manager'].apply_system_override.
             assert_called_once_with(
                {'configuration': {'root': {'@level': expected_level}}}))
            app_mocks['_run_nodetool_command'].assert_called_once_with(
                'setlogginglevel', 'root', expected_level)



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from mock import Mock
from mock import patch

from trove.common import pagination
from trove.guestagent.common import guestagent_utils
from trove.tests.unittests import trove_testtools


class TestGuestagentUtils(trove_testtools.TestCase):

    def test_update_dict(self):
        data = [{
            'dict': {}, 'update': {}, 'expected': {},
        }, {
            'dict': None, 'update': {}, 'expected': {},
        }, {
            'dict': {}, 'update': None, 'expected': {},
        }, {
            'dict': {}, 'update': None, 'expected': {},
        }, {
            'dict': None, 'update': {'name': 'Tom'},
            'expected': {'name': 'Tom'},
        }, {
            'dict': {}, 'update': {'name': 'Tom'},
            'expected': {'name': 'Tom'},
        }, {
            'dict': {'name': 'Tom'}, 'update': {},
            'expected': {'name': 'Tom'},
        }, {
            'dict': {'key1': 'value1',
                     'dict1': {'key1': 'value1', 'key2': 'value2'}},
            'update': {'key1': 'value1+',
                       'key2': 'value2',
                       'dict1': {'key3': 'value3'}},
            'expected': {'key1': 'value1+',
                         'key2': 'value2',
                         'dict1': {'key1': 'value1', 'key2': 'value2',
                                   'key3': 'value3'}},
        }, {
            'dict': {'d1': {'d2': {'d3': {'k1': 'v1'}}}},
            'update': {'d1': {'d2': {'d3': {'k2': 'v2'}}}},
            'expected': {'d1': {'d2': {'d3': {'k1': 'v1', 'k2': 'v2'}}}},
        }, {
            'dict': {'timeout': 0, 'save': [[900, 1], [300, 10]]},
            'update': {'save': [[300, 20], [60, 10000]]},
            'expected': {'timeout': 0,
                         'save': [[300, 20], [60, 10000]]},
        }, {
            'dict': {'rpc_address': '0.0.0.0',
                     'broadcast_rpc_address': '0.0.0.0',
                     'listen_address': '0.0.0.0',
                     'seed_provider': [{
                         'class_name':
                         'org.apache.cassandra.locator.SimpleSeedProvider',
                         'parameters': [{'seeds': '0.0.0.0'}]}]
                     },
            'update': {'rpc_address': '127.0.0.1',
                       'seed_provider': {'parameters': {
                           'seeds': '127.0.0.1'}}
                       },
            'expected': {'rpc_address': '127.0.0.1',
                         'broadcast_rpc_address': '0.0.0.0',
                         'listen_address': '0.0.0.0',
                         'seed_provider': [{
                             'class_name':
                             'org.apache.cassandra.locator.SimpleSeedProvider',
                             'parameters': [{'seeds': '127.0.0.1'}]}]
                         },
        }, {
            'dict': {'rpc_address': '127.0.0.1',
                     'broadcast_rpc_address': '0.0.0.0',
                     'listen_address': '0.0.0.0',
                     'seed_provider': [{
                         'class_name':
                         'org.apache.cassandra.locator.SimpleSeedProvider',
                         'parameters': [{'seeds': '0.0.0.0'}]}]
                     },
            'update': {'seed_provider':
                       [{'class_name':
                         'org.apache.cassandra.locator.SimpleSeedProvider'}]
                       },
            'expected': {'rpc_address': '127.0.0.1',
                         'broadcast_rpc_address': '0.0.0.0',
                         'listen_address': '0.0.0.0',
                         'seed_provider': [{
                             'class_name':
                             'org.apache.cassandra.locator.SimpleSeedProvider'
                         }]},
        }]
        count = 0
        for record in data:
            count += 1
            target = record['dict']
            update = record['update']
            expected = record['expected']
            result = guestagent_utils.update_dict(update, target)
            msg = 'Unexpected result for test %s' % str(count)
            self.assertEqual(expected, result, msg)

    def test_build_file_path(self):
        self.assertEqual(
            'base_dir/base_name',
            guestagent_utils.build_file_path('base_dir', 'base_name'))

        self.assertEqual(
            'base_dir/base_name.ext1',
            guestagent_utils.build_file_path('base_dir', 'base_name', 'ext1'))

        self.assertEqual(
            'base_dir/base_name.ext1.ext2',
            guestagent_utils.build_file_path(
                'base_dir', 'base_name', 'ext1', 'ext2'))

    def test_flatten_expand_dict(self):
        self._assert_flatten_expand_dict({}, {})
        self._assert_flatten_expand_dict({'ns1': 1}, {'ns1': 1})
        self._assert_flatten_expand_dict(
            {'ns1': {'ns2a': {'ns3a': True, 'ns3b': False}, 'ns2b': 10}},
            {'ns1.ns2a.ns3a': True, 'ns1.ns2a.ns3b': False, 'ns1.ns2b': 10})

    def _assert_flatten_expand_dict(self, nested_dict, flattened_dict):
        self.assertEqual(
            flattened_dict, guestagent_utils.flatten_dict(nested_dict))
        self.assertEqual(
            nested_dict, guestagent_utils.expand_dict(flattened_dict))

    def test_to_bytes(self):
        self.assertEqual('1024', guestagent_utils.to_bytes('1024'))
        self.assertEqual(1048576, guestagent_utils.to_bytes('1024K'))
        self.assertEqual(1073741824, guestagent_utils.to_bytes('1024M'))
        self.assertEqual(1099511627776, guestagent_utils.to_bytes('1024G'))
        self.assertEqual('1024T', guestagent_utils.to_bytes('1024T'))
        self.assertEqual(1024, guestagent_utils.to_bytes(1024))
        self.assertEqual('Hello!', guestagent_utils.to_bytes('Hello!'))
        self.assertEqual('', guestagent_utils.to_bytes(''))
        self.assertIsNone(guestagent_utils.to_bytes(None))

    @patch.object(pagination, 'paginate_object_list')
    def test_paginate_list(self, paginate_obj_mock):
        limit = Mock()
        marker = Mock()
        include_marker = Mock()
        test_list = [Mock(), Mock(), Mock()]
        guestagent_utils.paginate_list(
            test_list,
            limit=limit, marker=marker, include_marker=include_marker)
        paginate_obj_mock.assert_called_once_with(
            test_list, 'name',
            limit=limit, marker=marker, include_marker=include_marker)

    def test_serialize_list(self):
        test_list = [Mock(), Mock(), Mock()]
        with patch.object(guestagent_utils, 'paginate_list',
                          return_value=(test_list[:2], test_list[-2])
                          ) as paginate_lst_mock:
            _, next_name = guestagent_utils.serialize_list(test_list)
            paginate_lst_mock.assert_called_once_with(
                test_list,
                limit=None, marker=None, include_marker=False)
            for item in paginate_lst_mock.return_value[0]:
                item.serialize.assert_called_once_with()
            self.assertEqual(paginate_lst_mock.return_value[1], next_name)



# Copyright [2015] Hewlett-Packard Development Company, L.P.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mock import Mock, patch

from trove.common import exception
from trove.extensions.mgmt.clusters.models import MgmtCluster
from trove.extensions.mgmt.clusters.service import MgmtClusterController
from trove.tests.unittests import trove_testtools


class TestClusterController(trove_testtools.TestCase):
    def setUp(self):
        super(TestClusterController, self).setUp()

        self.context = trove_testtools.TroveTestContext(self)
        self.req = Mock()
        self.req.environ = Mock()
        self.req.environ.__getitem__ = Mock(return_value=self.context)

        mock_cluster1 = Mock()
        mock_cluster1.datastore_version.manager = 'vertica'
        mock_cluster1.instances = []
        mock_cluster1.instances_without_server = []
        mock_cluster2 = Mock()
        mock_cluster2.datastore_version.manager = 'vertica'
        mock_cluster2.instances = []
        mock_cluster2.instances_without_server = []
        self.mock_clusters = [mock_cluster1, mock_cluster2]

        self.controller = MgmtClusterController()

    def tearDown(self):
        super(TestClusterController, self).tearDown()

    def test_get_action_schema(self):
        body = {'do_stuff': {}}
        action_schema = Mock()
        action_schema.get = Mock()

        self.controller.get_action_schema(body, action_schema)
        action_schema.get.assert_called_with('do_stuff', {})

    @patch.object(MgmtCluster, 'load')
    def test_show_cluster(self, mock_cluster_load):
        tenant_id = Mock()
        id = Mock()
        mock_cluster_load.return_value = self.mock_clusters[0]

        self.controller.show(self.req, tenant_id, id)
        mock_cluster_load.assert_called_with(self.context, id)

    @patch.object(MgmtCluster, 'load_all')
    def test_index_cluster(self, mock_cluster_load_all):
        tenant_id = Mock()
        mock_cluster_load_all.return_value = self.mock_clusters

        self.controller.index(self.req, tenant_id)
        mock_cluster_load_all.assert_called_with(self.context, deleted=None)

    @patch.object(MgmtCluster, 'load')
    def test_controller_action_found(self, mock_cluster_load):
        body = {'reset-task': {}}
        tenant_id = Mock()
        id = Mock()
        mock_cluster_load.return_value = self.mock_clusters[0]

        result = self.controller.action(self.req, body, tenant_id, id)
        self.assertEqual(202, result.status)
        self.assertIsNotNone(result.data)

    def test_controller_no_body_action_found(self):
        tenant_id = Mock()
        id = Mock()

        self.assertRaisesRegex(
            exception.BadRequest, 'Invalid request body.',
            self.controller.action, self.req, None, tenant_id, id)

    @patch.object(MgmtCluster, 'load')
    def test_controller_invalid_action_found(self, mock_cluster_load):
        body = {'do_stuff': {}}
        tenant_id = Mock()
        id = Mock()
        mock_cluster_load.return_value = self.mock_clusters[0]

        self.assertRaisesRegex(
            exception.BadRequest, 'Invalid cluster action requested.',
            self.controller.action, self.req, body, tenant_id, id)



# Copyright 2015 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging

from trove.guestagent.datastore.mysql_common import service

LOG = logging.getLogger(__name__)


class KeepAliveConnection(service.BaseKeepAliveConnection):
    pass


class MySqlAppStatus(service.BaseMySqlAppStatus):
    pass


class LocalSqlClient(service.BaseLocalSqlClient):
    pass


class MySqlApp(service.BaseMySqlApp):
    def __init__(self, status):
        super(MySqlApp, self).__init__(status, LocalSqlClient,
                                       KeepAliveConnection)

    def _get_slave_status(self):
        with self.local_sql_client(self.get_engine()) as client:
            return client.execute('SHOW SLAVE STATUS').first()

    def _get_master_UUID(self):
        slave_status = self._get_slave_status()
        return slave_status and slave_status['Master_UUID'] or None

    def _get_gtid_executed(self):
        with self.local_sql_client(self.get_engine()) as client:
            return client.execute('SELECT @@global.gtid_executed').first()[0]

    def get_last_txn(self):
        master_UUID = self._get_master_UUID()
        last_txn_id = '0'
        gtid_executed = self._get_gtid_executed()
        for gtid_set in gtid_executed.split(','):
            uuid_set = gtid_set.split(':')
            if uuid_set[0] == master_UUID:
                last_txn_id = uuid_set[-1].split('-')[-1]
                break
        return master_UUID, int(last_txn_id)

    def get_latest_txn_id(self):
        LOG.info("Retrieving latest txn id.")
        return self._get_gtid_executed()

    def wait_for_txn(self, txn):
        LOG.info("Waiting on txn '%s'.", txn)
        with self.local_sql_client(self.get_engine()) as client:
            client.execute("SELECT WAIT_UNTIL_SQL_THREAD_AFTER_GTIDS('%s')"
                           % txn)


class MySqlRootAccess(service.BaseMySqlRootAccess):
    def __init__(self):
        super(MySqlRootAccess, self).__init__(LocalSqlClient,
                                              MySqlApp(MySqlAppStatus.get()))


class MySqlAdmin(service.BaseMySqlAdmin):
    def __init__(self):
        super(MySqlAdmin, self).__init__(LocalSqlClient, MySqlRootAccess(),
                                         MySqlApp)



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import glob
import os
import re
import tempfile

from oslo_log import log as logging
import pexpect

from trove.common import cfg
from trove.common import exception
from trove.common import utils
from trove.guestagent.common import operating_system
from trove.guestagent.common.operating_system import FileMode
import trove.guestagent.datastore.mysql.service as dbaas
from trove.guestagent.strategies.restore import base

LOG = logging.getLogger(__name__)


class MySQLRestoreMixin(object):
    """Common utils for restoring MySQL databases."""
    RESET_ROOT_RETRY_TIMEOUT = 100
    RESET_ROOT_SLEEP_INTERVAL = 10

    RESET_ROOT_MYSQL_COMMANDS = ("SET PASSWORD FOR "
                                 "'root'@'localhost'='';")
    # This is a suffix MySQL appends to the file name given in
    # the '--log-error' startup parameter.
    _ERROR_LOG_SUFFIX = '.err'
    _ERROR_MESSAGE_PATTERN = re.compile(b"ERROR")

    def mysql_is_running(self):
        try:
            utils.execute_with_timeout("/usr/bin/mysqladmin", "ping")
            LOG.debug("MySQL is up and running.")
            return True
        except exception.ProcessExecutionError:
            LOG.debug("MySQL is not running.")
            return False

    def mysql_is_not_running(self):
        try:
            utils.execute_with_timeout("/usr/bin/pgrep", "mysqld")
            LOG.debug("MySQL is still running.")
            return False
        except exception.ProcessExecutionError:
            LOG.debug("MySQL is not running.")
            return True

    def poll_until_then_raise(self, event, exc):
        try:
            utils.poll_until(event,
                             sleep_time=self.RESET_ROOT_SLEEP_INTERVAL,
                             time_out=self.RESET_ROOT_RETRY_TIMEOUT)
        except exception.PollTimeOut:
            raise exc

    def _start_mysqld_safe_with_init_file(self, init_file, err_log_file):
        # This directory is added and removed by the mysql systemd service
        # as the database is started and stopped. The restore operation
        # takes place when the database is stopped, so the directory does
        # not exist, but it is assumed to exist by the mysqld_safe command
        # which starts the database. This command used to create this
        # directory if it didn't exist, but it was changed recently to
        # simply fail in this case.
        run_dir = "/var/run/mysqld"
        if not os.path.exists(run_dir):
            utils.execute("mkdir", run_dir,
                          run_as_root=True, root_helper="sudo")
        utils.execute("chown", "mysql:mysql", run_dir, err_log_file.name,
                      init_file.name, run_as_root=True, root_helper="sudo")
        command_mysql_safe = ("sudo mysqld_safe"
                              " --init-file=%s"
                              " --log-error=%s" %
                              (init_file.name, err_log_file.name))
        LOG.debug("Spawning: %s" % command_mysql_safe)
        child = pexpect.spawn(command_mysql_safe)
        try:
            index = child.expect(['Starting mysqld daemon'])
            if index == 0:
                LOG.info("Starting MySQL")
        except pexpect.TIMEOUT:
            LOG.exception("Got a timeout launching mysqld_safe")
        finally:
            # There is a race condition here where we kill mysqld before
            # the init file been executed. We need to ensure mysqld is up.
            #
            # mysqld_safe will start even if init-file statement(s) fail.
            # We therefore also check for errors in the log file.
            self.poll_until_then_raise(
                self.mysql_is_running,
                base.RestoreError("Reset root password failed:"
                                  " mysqld did not start!"))
            first_err_message = self._find_first_error_message(err_log_file)
            if first_err_message:
                raise base.RestoreError("Reset root password failed: %s"
                                        % first_err_message)

            LOG.info("Root password reset successfully.")
            LOG.debug("Cleaning up the temp mysqld process.")
            utils.execute_with_timeout("mysqladmin", "-uroot",
                                       "--protocol=tcp", "shutdown")
            LOG.debug("Polling for shutdown to complete.")
            try:
                utils.poll_until(self.mysql_is_not_running,
                                 sleep_time=self.RESET_ROOT_SLEEP_INTERVAL,
                                 time_out=self.RESET_ROOT_RETRY_TIMEOUT)
                LOG.debug("Database successfully shutdown")
            except exception.PollTimeOut:
                LOG.debug("Timeout shutting down database "
                          "- performing killall on mysqld_safe.")
                utils.execute_with_timeout("killall", "mysqld_safe",
                                           root_helper="sudo",
                                           run_as_root=True)
                self.poll_until_then_raise(
                    self.mysql_is_not_running,
                    base.RestoreError("Reset root password failed: "
                                      "mysqld did not stop!"))

    def reset_root_password(self):
        """Reset the password of the localhost root account used by Trove
        for initial datastore configuration.
        """

        try:
            # Do not attempt to delete these files as the 'trove' user.
            # The process writing into it may have assumed its ownership.
            # Only owners can delete temporary files (restricted deletion).
            init_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
            operating_system.write_file(init_file.name,
                                        self.RESET_ROOT_MYSQL_COMMANDS)
            operating_system.chmod(init_file.name, FileMode.ADD_READ_ALL,
                                   as_root=True)
            err_log_file = tempfile.NamedTemporaryFile(
                suffix=self._ERROR_LOG_SUFFIX,
                delete=False)
            self._start_mysqld_safe_with_init_file(init_file, err_log_file)
        finally:
            init_file.close()
            err_log_file.close()
            operating_system.remove(
                init_file.name, force=True, as_root=True)
            operating_system.remove(
                err_log_file.name, force=True, as_root=True)

    def _find_first_error_message(self, fp):
        if self._is_non_zero_file(fp):
            return self._find_first_pattern_match(
                fp, self._ERROR_MESSAGE_PATTERN)
        return None

    def _is_non_zero_file(self, fp):
        file_path = fp.name
        return os.path.isfile(file_path) and (os.path.getsize(file_path) > 0)

    def _find_first_pattern_match(self, fp, pattern):
        for line in fp:
            if pattern.match(line):
                return line
        return None


class MySQLDump(base.RestoreRunner, MySQLRestoreMixin):
    """Implementation of Restore Strategy for MySQLDump."""
    __strategy_name__ = 'mysqldump'
    base_restore_cmd = 'sudo mysql'


class InnoBackupEx(base.RestoreRunner, MySQLRestoreMixin):
    """Implementation of Restore Strategy for InnoBackupEx."""
    __strategy_name__ = 'innobackupex'
    base_restore_cmd = ('sudo xbstream -x -C %(restore_location)s'
                        ' 2>/tmp/xbstream_extract.log')
    base_prepare_cmd = ('sudo innobackupex'
                        ' --defaults-file=%(restore_location)s/backup-my.cnf'
                        ' --ibbackup=xtrabackup'
                        ' --apply-log'
                        ' %(restore_location)s'
                        ' 2>/tmp/innoprepare.log')

    def __init__(self, *args, **kwargs):
        self._app = None
        super(InnoBackupEx, self).__init__(*args, **kwargs)
        self.prepare_cmd = self.base_prepare_cmd % kwargs
        self.prep_retcode = None

    @property
    def app(self):
        if self._app is None:
            self._app = self._build_app()
        return self._app

    def _build_app(self):
        return dbaas.MySqlApp(dbaas.MySqlAppStatus.get())

    def pre_restore(self):
        self.app.stop_db()
        LOG.debug("Cleaning out restore location: %s.", self.restore_location)
        operating_system.chmod(self.restore_location, FileMode.SET_FULL,
                               as_root=True)
        utils.clean_out(self.restore_location)

    def _run_prepare(self):
        LOG.info("Running innobackupex prepare: %s.", self.prepare_cmd)
        self.prep_retcode = utils.execute(self.prepare_cmd, shell=True)
        LOG.info("Innobackupex prepare finished successfully.")

    def post_restore(self):
        self._run_prepare()
        operating_system.chown(self.restore_location, 'mysql', None,
                               force=True, as_root=True)
        self._delete_old_binlogs()
        self.reset_root_password()
        self.app.start_mysql()

    def _delete_old_binlogs(self):
        files = glob.glob(os.path.join(self.restore_location, "ib_logfile*"))
        for f in files:
            os.unlink(f)

    def check_process(self):
        """Check whether xbstream restore is successful."""
        # We first check the restore process exits with 0, however
        # xbstream has a bug for creating new files:
        # https://jira.percona.com/browse/PXB-1542
        # So we also check the stderr with ignorance of some known
        # non-error log lines. Currently we only need to ignore:
        # "encryption: using gcrypt x.x.x"
        # After PXB-1542 is fixed, we could just check the exit status.
        LOG.debug('Checking return code of xbstream restore process.')
        return_code = self.process.wait()
        if return_code != 0:
            LOG.error('xbstream exited with %s', return_code)
            return False

        LOG.debug('Checking xbstream restore process stderr output.')
        IGNORE_LINES = [
            'encryption: using gcrypt ',
            'sudo: unable to resolve host ',
        ]
        with open('/tmp/xbstream_extract.log', 'r') as xbstream_log:
            for line in xbstream_log:
                # Ignore empty lines
                if not line.strip():
                    continue

                # Ignore known non-error log lines
                check_ignorance = [line.startswith(non_err)
                                   for non_err in IGNORE_LINES]
                if any(check_ignorance):
                    continue
                else:
                    LOG.error('xbstream restore failed with: %s',
                              line.rstrip('\n'))
                    return False

        return True


class InnoBackupExIncremental(InnoBackupEx):
    __strategy_name__ = 'innobackupexincremental'
    incremental_prep = ('sudo innobackupex'
                        ' --defaults-file=%(restore_location)s/backup-my.cnf'
                        ' --ibbackup=xtrabackup'
                        ' --apply-log'
                        ' --redo-only'
                        ' %(restore_location)s'
                        ' %(incremental_args)s'
                        ' 2>/tmp/innoprepare.log')

    def __init__(self, *args, **kwargs):
        super(InnoBackupExIncremental, self).__init__(*args, **kwargs)
        self.restore_location = kwargs.get('restore_location')
        self.content_length = 0

    def _incremental_restore_cmd(self, incremental_dir):
        """Return a command for a restore with a incremental location."""
        args = {'restore_location': incremental_dir}
        return (self.decrypt_cmd +
                self.unzip_cmd +
                (self.base_restore_cmd % args))

    def _incremental_prepare_cmd(self, incremental_dir):
        if incremental_dir is not None:
            incremental_arg = '--incremental-dir=%s' % incremental_dir
        else:
            incremental_arg = ''

        args = {
            'restore_location': self.restore_location,
            'incremental_args': incremental_arg,
        }

        return self.incremental_prep % args

    def _incremental_prepare(self, incremental_dir):
        prepare_cmd = self._incremental_prepare_cmd(incremental_dir)
        LOG.debug("Running innobackupex prepare: %s.", prepare_cmd)
        utils.execute(prepare_cmd, shell=True)
        LOG.debug("Innobackupex prepare finished successfully.")

    def _incremental_restore(self, location, checksum):
        """Recursively apply backups from all parents.

        If we are the parent then we restore to the restore_location and
        we apply the logs to the restore_location only.

        Otherwise if we are an incremental we restore to a subfolder to
        prevent stomping on the full restore data. Then we run apply log
        with the '--incremental-dir' flag
        """
        metadata = self.storage.load_metadata(location, checksum)
        incremental_dir = None
        if 'parent_location' in metadata:
            LOG.info("Restoring parent: %(parent_location)s"
                     " checksum: %(parent_checksum)s.", metadata)
            parent_location = metadata['parent_location']
            parent_checksum = metadata['parent_checksum']
            # Restore parents recursively so backup are applied sequentially
            self._incremental_restore(parent_location, parent_checksum)
            # for *this* backup set the incremental_dir
            # just use the checksum for the incremental path as it is
            # sufficiently unique /var/lib/mysql/<checksum>
            incremental_dir = os.path.join(
                cfg.get_configuration_property('mount_point'), checksum)
            operating_system.create_directory(incremental_dir, as_root=True)
            command = self._incremental_restore_cmd(incremental_dir)
        else:
            # The parent (full backup) use the same command from InnobackupEx
            # super class and do not set an incremental_dir.
            command = self.restore_cmd

        self.content_length += self._unpack(location, checksum, command)
        self._incremental_prepare(incremental_dir)

        # Delete unpacked incremental backup metadata
        if incremental_dir:
            operating_system.remove(incremental_dir, force=True, as_root=True)

    def _run_restore(self):
        """Run incremental restore.

        First grab all parents and prepare them with '--redo-only'. After
        all backups are restored the super class InnoBackupEx post_restore
        method is called to do the final prepare with '--apply-log'
        """
        self._incremental_restore(self.location, self.checksum)
        return self.content_length



# Copyright 2016 Tesora, Inc
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.common import wsgi
from trove.volume_type import models
from trove.volume_type import views


class VolumeTypesController(wsgi.Controller):
    """A controller for the Cinder Volume Types functionality."""

    def show(self, req, tenant_id, id):
        """Return a single volume type."""
        context = req.environ[wsgi.CONTEXT_KEY]
        volume_type = models.VolumeType.load(id, context=context)
        return wsgi.Result(views.VolumeTypeView(volume_type, req).data(), 200)

    def index(self, req, tenant_id):
        """Return all volume types."""
        context = req.environ[wsgi.CONTEXT_KEY]
        volume_types = models.VolumeTypes(context=context)
        return wsgi.Result(views.VolumeTypesView(volume_types,
                                                 req).data(), 200)



# Copyright 2015 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_utils import importutils
from trove.guestagent.datastore.mysql_common import manager


MYSQL_APP = ("trove.guestagent.datastore.experimental.percona.service."
             "MySqlApp")
MYSQL_APP_STATUS = ("trove.guestagent.datastore.experimental.percona.service."
                    "MySqlAppStatus")
MYSQL_ADMIN = ("trove.guestagent.datastore.experimental.percona.service."
               "MySqlAdmin")


class Manager(manager.MySqlManager):

    def __init__(self):
        mysql_app = importutils.import_class(MYSQL_APP)
        mysql_app_status = importutils.import_class(MYSQL_APP_STATUS)
        mysql_admin = importutils.import_class(MYSQL_ADMIN)

        super(Manager, self).__init__(mysql_app, mysql_app_status, mysql_admin)



# Copyright (c) 2014 eBay Software Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import json

from oslo_log import log as logging

from trove.common import exception
from trove.common import utils
from trove.guestagent.common import operating_system
from trove.guestagent.datastore.experimental.couchbase import service
from trove.guestagent.datastore.experimental.couchbase import system
from trove.guestagent.strategies.backup import base


LOG = logging.getLogger(__name__)
OUTFILE = '/tmp' + system.BUCKETS_JSON


class CbBackup(base.BackupRunner):
    """
    Implementation of Backup Strategy for Couchbase.
    """
    __strategy_name__ = 'cbbackup'

    pre_backup_commands = [
        ['rm', '-rf', system.COUCHBASE_DUMP_DIR],
        ['mkdir', '-p', system.COUCHBASE_DUMP_DIR],
    ]

    post_backup_commands = [
        ['rm', '-rf', system.COUCHBASE_DUMP_DIR],
    ]

    @property
    def cmd(self):
        """
        Creates backup dump dir, tars it up, and encrypts it.
        """
        cmd = 'tar cpPf - ' + system.COUCHBASE_DUMP_DIR
        return cmd + self.zip_cmd + self.encrypt_cmd

    def _save_buckets_config(self, password):
        url = system.COUCHBASE_REST_API + '/pools/default/buckets'
        utils.execute_with_timeout('curl -u root:' + password +
                                   ' ' + url + ' > ' + OUTFILE,
                                   shell=True, timeout=300)

    def _backup(self, password):
        utils.execute_with_timeout('/opt/couchbase/bin/cbbackup',
                                   system.COUCHBASE_REST_API,
                                   system.COUCHBASE_DUMP_DIR,
                                   '-u', 'root', '-p', password,
                                   timeout=600, run_as_root=True,
                                   root_helper='sudo')

    def _run_pre_backup(self):
        try:
            for cmd in self.pre_backup_commands:
                utils.execute_with_timeout(*cmd)
            root = service.CouchbaseRootAccess()
            pw = root.get_password()
            self._save_buckets_config(pw)
            with open(OUTFILE, "r") as f:
                out = f.read()
                if out != "[]":
                    d = json.loads(out)
                    all_memcached = True
                    for i in range(len(d)):
                        bucket_type = d[i]["bucketType"]
                        if bucket_type != "memcached":
                            all_memcached = False
                            break
                    if not all_memcached:
                        self._backup(pw)
                    else:
                        LOG.info("All buckets are memcached.  "
                                 "Skipping backup.")
            operating_system.move(OUTFILE, system.COUCHBASE_DUMP_DIR)
            if pw != "password":
                # Not default password, backup generated root password
                operating_system.copy(system.pwd_file,
                                      system.COUCHBASE_DUMP_DIR,
                                      preserve=True, as_root=True)
        except exception.ProcessExecutionError:
            LOG.exception("Error during pre-backup phase.")
            raise

    def _run_post_backup(self):
        try:
            for cmd in self.post_backup_commands:
                utils.execute_with_timeout(*cmd)
        except exception.ProcessExecutionError:
            LOG.exception("Error during post-backup phase.")
            raise
